#####################################################################
#:TriggerCQCC.pm
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2013. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME

CQCC::TriggerCQCC - Class that defines user interactions with the 
ClearCase/ClearQuest integration.

=head1 CLASS DESCRIPTION

This class defines the overall behavior of the trigger application
drawing upon the capabilities of the lower level classes to handle
user interaction and communications with ClearCase and ClearQuest.
It relies upon its superclass, TriggerBase, for methods to handle
command line argument processing, initialization, menu interaction, 
and cache management.

The trigger is activated by ClearCase operations, each of which has
its own entry method; for example, when the user does a checkin, then
the Op_checkin method is invoked.

Depending on the operation, the main menu is presented to the user for
runtime interaction. The menu in turn calls either Dialog methods,
which interact with the user, or Action methods, which carry out an
action and possibly exit from the trigger.

NOTE: Anyone interested in understanding how the overall trigger works
or who is contemplating customizing the behavior of the trigger should
start with a review of this class and how its methods work. Most
changes in visible behavior of the trigger will involve changes in
this class. For additional source code documentation, contact Rational
Customer Support. 

If local changes are needed, do NOT make them directly to the
TriggerCQCC source code: make a copy of the MyTrigger.pm template
class, rename it, and make your changes there by overriding or
extending the TriggerCQCC methods (see MyTrigger.pm for more
details). Doing this will facilitate upgrading to later releases of
the CQCC trigger from Rational and provide a fallback to the released
trigger for working with Rational Customer Support. 

SUPPORT POLICY: 

Rational Support supports the new integration as provided and local
configuration changes made using the parameters in the configuration
file (config.pl). 

Rational does NOT support source code changes to the integration even
though it is provided in source code form as Perl scripts. Internal
source code documentation will be provided on request to enable local
sites to make judicious changes if needed but Rational Customer
Support can not and will not support local changes of this sort.

=head1 METHODS

=cut

######################################################################
#::Package CQCC::TriggerCQCC
######################################################################

package CQCC::TriggerCQCC;
use Exporter;
use CQCC::TriggerBase;
@ISA = ('Exporter', "CQCC::TriggerBase" );
@EXPORT = qw(&TriggerMain);

use strict;

# Note: These classes are all the base classes that the trigger
# needs directly. At runtime, subclasses may actually be loaded
# and used instead. For example, CQuest is specified here but
# in practice either CQWeb or CQPerlAPI would be used.

use CQCC::CQuest;
use CQCC::CCase;
use CQCC::IOBase;
use CQCC::Parm;
use CQCC::AssocList;
use CQCC::Timer;
use CQCC::CQSchema;


sub Version { return '2.2' };

######################################################################
#::Class Globals
######################################################################

my $addOp = "<+>";
my $delOp = "<->";
my $perl5006 = ($] ge "5.006" ? 1 : 0);

######################################################################
#::function TriggerMain

=over 6

=item C<TriggerMain()>

This is the main entry subroutine for the application. It will call
methods to evaluate the command line and environment settings, call
the method with the same name as the "operation", then cleanup.

=back

=cut

######################################################################

sub TriggerMain {
    # Make sure we configure IOBaseInterface before it might be needed. 

    if ($ENV{ATRIA_WEB_GUI} && ! defined $ENV{ATRIA_WEB_CLEARPROMPT}) {
       &IOMessage("FATAL", ["TCQCC_E_WEB_NOT_SUPPORTED"]);
    }

        my $ioClass = "CQCC::IOBase";
    if (! &ParmDefined("IOBaseInterface", "CONFIG")) {
        if ($ENV{ATRIA_FORCE_GUI} || $ENV{ATRIA_WEB_GUI} ||
                &GetParm("CQCC_GUI_ENABLE") eq "ALWAYS") {
            $ioClass = "CQCC::IOClearPrompt";
            if (&GetParm("CQCC_GUI_ENABLE") ne "OFF" &&
                ! defined $ENV{ATRIA_WEB_GUI}) {
                my $TKavailable = 0;
                if ($perl5006) {
                    for (@INC) {
                        if (-e $_ . "/" . "Tk.pm") {
                            $TKavailable = 1;
                            last;
                        }
                    }
                    if ($TKavailable) { 
                        $ioClass = "CQCC::IOPerlTK";
                    }
                }
             }
             &SetConfigParm("IOBaseInterface", $ioClass);
         }
    }
 
    # Complete IOBase initialization with latest msgcat and conversion
    # settings. 

    CQCC::IOBase::IOCompleteSetup();  

    # Preprocess the arguments in case one of the command line
    # parameter changes affects the TriggerInterface parameter.

    &CQCC::TriggerBase::PreprocessArguments();
    my $trigger = &IOClassFactory("TriggerInterface");
    $trigger->Initialize();
    $trigger->CleanCacheDir(&IOPathname("PARM","."),
                            &GetParm("CM_STALE_HRS"),
                            &GetParm("CM_CLEAN_HRS"));


    $trigger->MenuBuild();

    # This will call the appropriate operation method based on the
    # CLEARCASE_OPKIND variable (i.e. "checkin" calls "Op_checkin")
    # or the -op command line argument. 

    $trigger->DispatchOperation();
}

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class
parameters using CQCC::Parm objects. Current parameters include:

=over 4

=item C<CQCC_MULTIPLE_ASSOCS>

This switch selects whether the user can make only one association
('FALSE') or multiple associations ('TRUE', default) per operation. 

=item C<CQCC_ASSOC_BATCH_ENABLE>

This parameter enables logging ClearQuest association transactions so
that they can be processed together at a later time for improved
performance when handling multiple files in the same operations. The
operations are logged to <param-dir>/.cqcc_assoc_batch, where param-dir 
is the user's home directory (UNIX) or Windows application data 
directory.

The log is automatically processed at the end of each single operation
or at the end of a batch operation. A batch is defined as either a set
of files used in one ClearCase operation such as a multiple-checkout,
or as defined by the user using the CQCC_ASSOC_BATCH_SERIES
environment variable.

An additional mode (AUTOFLUSH) will cause batches to be processed for
each file instead of waiting for the end of the series. This mode 
eliminates the performance benefits of batching but still provides the 
reliability of saving the pending CQ transaction to a file before trying 
to perform the CQ operation. In case the operation fails, the transaction
will be re-tried with the next Clearcase operation. This mode also provides
a way to gradually wean a site off of using batching, since turning off
batching could otherwise leave unposted transaction files. 

=item C<CQCC_ASSOC_BATCHFILE (Unpublished)>

This parameter provides the filename for logging ClearQuest
association transactions that is enabled by CQCC_ASSOC_BATCH_ENABLE.
By default, it is set to "STANDARD" which sets the name automatically
set to <param-dir>/.cqcc_assoc_batch, where param-dir is the same 
directory where the integration stores its cache and CQ login 
information. Any other non-empty value will be treated as a specific
filename to be used instead. 

=item C<CQCC_ASSOC_BATCH_AUTORUN (Unpublished)>

This parameter enables the integration to automatically process the
logged ClearQuest association information at the end of a file series
(as marked by ClearCase) or at the end of a user defined series (see
CQCC_ASSOC_BATCH_SERIES). Set to TRUE (default) or FALSE (user
expected to use "cqcc_launch -op batch <filename>" mechanism to
initiate applying the logged transactions.

=item C<CQCC_ASSOC_BATCH_CONFIRM>

This parameter enables the display of a message to the end user when a
batch has been successfully posted. Since the batch won't be
automatically applied right away if the last checkin fails, this
provides a visual cue to the user that they may want to force the
batch to fire manually. There are three settings: "OFF" - nothing is
reported; "MULTIPLE" - displayed when multiple files are being processed
at once; and "ALWAYS" - always display the message for single or multiple
files being processed. 

=item C<CQCC_ASSOC_BATCH_SERIES>

This parameter allows the user to identify a series of ClearCase
transactions to be treated as a file series to control when the batch
autorun feature is activated (this doesn't affect the ApplyToAll
options presented to the user). Normally disabled ("FALSE") by
default, it is set to "TRUE" to indicate that a user-defined series 
is underway and to suppress normal ClearCase series-end batch
processing. When set back to "FALSE", the next ClearCase series-end
operation (end of a series or an individual version checkin/out) will
trigger automatic batch processing as usual.

=item C<CQCC_AUTO_ASSOCIATE>

Use this option to specify one or more change requests that you want
to associate with files on checkout and checkin operations. See the
README file or online help for details. 

=item C<CQCC_AUTO_ASSOCIATE_ENABLE>

Use this option to specify whether users may use the
CQCC_AUTO_ASSOCIATE option. See config.pl for more details.

=item C<CQCC_COMMENT_PATTERN>

Use this option to specify a comment pattern that will allow users to
make associations by entering a checkin or checkout comment that
matches the pattern. See config.pl for more details. 

=item C<CQCC_CQWEB_ONLY>

This switch is used to force using the CQWeb interface even when
ClearQuest and CQPerl is found on the client machine. It is set to
'FALSE' by default; set it to 'TRUE' to always use CQWeb.

=item C<CQCC_GUI_ENABLE>

This switch is used to enable or disable the Perl/TK GUI interface.
If 'ON' (default) then it will be presented when possible (Perl/TK
support is available and ClearCase Explorer or XClearCase is in
use). If 'ALWAYS', then the GUI will be presented for command line
access as well, when possible. If 'OFF', then the GUI is never
presented and Perl/TK is not loaded.

=item C<CQCC_POSTCHECKIN_COMMIT>

This switch enables the integration to delay committing associations
on checkin until after the ClearCase Checkin operation has completed.

This avoids problems caused in cases where the ClearCase operation 
has failed. For example, checking in identical files without the
-identical option won't succeed; if the failed files are then checked
out the associations committed earlier are incorrect. Using this
option avoids this problem. This option supercedes
CQCC_POSTCHECKIN_VERIFY. Using this option may require an extra
ClearQuest login on the post-op trigger; using CQCC_ASSOC_BATCH_ENABLE 
may minimize this cost. 

=item C<CQCC_POSTCHECKIN_VERIFY>

This switch enables a post-checkin validation that the object IDs
for the version file that are being stored in ClearQuest are correct.
This may be needed because the object IDs can change after the version
is checked in. If they do have changed, they will be updated if this 
switch is set to 'TRUE'. If it is set to 'FALSE' (default), check is 
not made. 

=item C<CQCC_QUERY_ENABLE>

This switch controls whether the user can select different queries
while searching for associations. If 'LOCAL' (default), then only
queries defined locally in the configuration file are presented. 
If 'CQ' then only ClearQuest named queries will be shown. If 'BOTH'
then both LOCAL and CQ queries are displayed. If 'OFF' then no choice
is permitted; the GUI's Browse button is disabled and the QueryName
menu option is hidden. Query names displayed are subject to the 
CQCC_QUERY_FILTER for finer control. 

=item C<CQCC_QUERY_FILTER>

This switch provides an optional filter to control which query names
will be presented to the user to choose from. The string is applied
in a Perl pattern match if it is not empty. For example, to only
allow queries that include "Public Queries", set the filter string
to "Public Queries". By default the string is empty. 

=item C<CQCC_CQWEB_VERSION>

There are two versions of CQWeb server protocols, CQWeb ASP and CQWeb
Java. This parameter controls which protocol the integration is
using. If it is unset or set to "2.0" (default) then CQWeb Java is
assumed; if set to a version number such as "1.0", then CQWeb ASP (old
version) is assumed instead.

=item C<CQCC_RESTRICTIONS_TIMEOUT>

This controls the number of seconds that a restrictions check can be
reused during CQCC_ASSOC_BATCH log processing. After the time is
exceeded, a new CQ login and restrictions will automatically be 
performed to make sure the selected associations are still valid. 
The Value ranges from 0 (disabled) to 1200 (20 minutes), with a default 
of 300 (5 minutes).

=item C<CQCC_MKHLINK_CCXPN>

This switch provides an optional method when creating an hlink on windows.
This feature should be enabled only if the mkhlink fails.  This is
used for windows only.  This option may be needed when accessing UNIX
VOBs via the MVFS drive and other non integration preop triggers are
firing.  Turning on this option may lead to the wrong file being
linked if there are multiple checkouts of that file.  A safer option
is to mount the view instead of using the MVFS drive or just turning
this option on as needed by setting the environment variable.

=item C<CQCC_CQWEB_ASSOC_RETRY>

This option is used to specify whether we perform a retry (and retry times)
when the integration fails making associations with CQ record. This option 
is only effective when CC and CQ use the web connection. The Value ranges 
from 0 (disabled) to 7 (retry 7 times), with a default of 0 (disabled).

=item C<CACHE_*>

There are a number of internal parameters maintained by this class
that hold the values of runtime values from the previous trigger
invocation. They are not externally visible or settable. 

=back

=back

=cut

######################################################################

sub ClassInit {
    # Environment Parameters

    if (&ParmDefined("CQCC_MULTIPLE_ASSOCS")) {
        return;
    }

    # Configuration Parameters

    &SetParmDefaults(PCLASS=>"CQCC::TriggerCQCC");

    &DefineParm(NAME=>"CQCC_MULTIPLE_ASSOCS", SRC_CFG=>1,
                DEFAULT=>"TRUE",
                RANGE_LIST => ["FALSE", "TRUE"]);

    &DefineParm(NAME=>"CQCC_ASSOC_BATCH_ENABLE", SRC_CFG=>1, SRC_EV=>1,
                DEFAULT=>"FALSE", RANGE_LIST => ["FALSE", "TRUE", "AUTOFLUSH"]);
    # BATCHFILE and AUTORUN are hidden from casual users
    &DefineParm(NAME=>"CQCC_ASSOC_BATCHFILE", SRC_CFG=>1, SRC_EV=>1, 
                DEFAULT=>"STANDARD");
    &DefineParm(NAME=>"CQCC_ASSOC_BATCH_AUTORUN", SRC_CFG=>1, SRC_EV=>1,
                DEFAULT=>"TRUE", RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"CQCC_ASSOC_BATCH_CONFIRM", SRC_CFG=>1, SRC_EV=>1,
                DEFAULT=>"OFF", RANGE_LIST => ["OFF", "MULTIPLE", "ALWAYS"]);
    &DefineParm(NAME=>"CQCC_ASSOC_BATCH_SERIES", SRC_EV=>1, 
                DEFAULT=>"FALSE", RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"CQCC_AUTO_ASSOCIATE_ENABLE", SRC_CFG=>1,
                DEFAULT=>"TRUE",
                RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"CQCC_AUTO_ASSOCIATE", SRC_EV=>1,
                DEFAULT=>"");
    &DefineParm(NAME=>"CQCC_COMMENT_PATTERN", SRC_CFG=>1,
                DEFAULT=>"");
    &DefineParm(NAME=>"CQCC_CQWEB_ONLY", SRC_CFG=>1,
                DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"CQCC_GUI_ENABLE", SRC_CFG=>1, SRC_EV=>1,
                DEFAULT=>"ON",
                RANGE_LIST => ["OFF", "ON", "ALWAYS" ]);
    &DefineParm(NAME=>"CQCC_POSTCHECKIN_COMMIT", SRC_CFG=>1,
                DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"CQCC_POSTCHECKIN_VERIFY", SRC_CFG=>1,
                DEFAULT=>"FALSE", 
                RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"CQCC_QUERY_ENABLE", SRC_CFG=>1,
                DEFAULT=>"LOCAL",
                RANGE_LIST => ["OFF", "LOCAL", "CQ", "BOTH" ]);
    &DefineParm(NAME=>"CQCC_QUERY_FILTER", SRC_CFG=>1,
                DEFAULT=>"");
    &DefineParm(NAME=>"CQCC_RESTRICTIONS_TIMEOUT", SRC_CFG=>1, SRC_EV=>1,
                DEFAULT=>300, RANGE_MIN=>0, RANGE_MAX=>1200);
    &DefineParm(NAME=>"CQCC_CQWEB_VERSION", DEFAULT=>'2.0',
                SRC_CFG=>1, SRC_EV=>1);
    &DefineParm(NAME=>"CQCC_MKHLINK_CCXPN", SRC_CFG=>1, SRC_EV=>1,
                DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"CQCC_CQWEB_ASSOC_RETRY", SRC_EV=>1,
                DEFAULT=>0,
                RANGE_MIN=>0, RANGE_MAX=>7);
    
    # Internal state variables, cached between trigger calls

    &DefineParm(NAME=>"DesiredAssocList", SRC_INT=>1);

    &DefineParm(NAME=>"CACHE_DESIRED_ASSOCS",
                SRC_INT=>1, SRC_CACHE=>1);
    &DefineParm(NAME=>"CACHE_APPLY_TO_ALL", DEFAULT=>0,
                SRC_INT=>1, SRC_CACHE=>1);
    &DefineParm(NAME=>"CACHE_RECENT_ASSOCS",
                SRC_INT=>1, SRC_CACHE=>1);
    &DefineParm(NAME=>"CACHE_SERIES", DEFAULT=>0,
                SRC_INT=>1, SRC_CACHE=>1);
    &DefineParm(NAME=>"CACHE_VOLIST", DEFAULT=>"",
                SRC_INT=>1, SRC_CACHE=>1);


    # Special option for support to perform extended diagnosis only.
    # Never enable this unless you have been instructed by support to do so.
    &DefineParm(NAME=>"CQCC_ASSOC_BATCH_FILE_OVERWRITE", SRC_CFG=>1,
                DEFAULT=>"TRUE",
                RANGE_LIST => ["FALSE", "TRUE"]);

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::method ConfigureClasses

=over 6

=item C<ConfigureClasses()>

This method decides which component classes to use when initializing
the trigger based on the current environment. This may be overridden
to use a different set of criteria.

Controlling which classes are used allows the administrator to decide
which standard interface is used for ClearCase (Process or PerlApi)
and ClearQuest (Web or PerlAPI) or which local subclass is used to
support local extensions or overrides.  Interfaces only supported on
certain Perls or platforms can be added on here without touching lower
level code.

=back

=cut

######################################################################

sub ConfigureClasses {
    my $self = shift;

    # Note: IOBaseInterface is configured directly by TriggerMain

    if ($self->{CLASSNAME} eq "CQCC::GDTrigger") {
       &IOMessage("FATAL", ["TCQCC_E_GDTRIGGER_INCOMPATIBLE"]);
    }

    # When CCPerlAPI is available, use that - need a test first
    if (! &ParmDefined("ClearCaseInterface", "CONFIG")) {
        &SetConfigParm("ClearCaseInterface", "CQCC::CCase");
    } 

    # Assume that if we are running cqperl (version > 5.6) then
    # we should be using CQPerlAPI

    if (! &ParmDefined("ClearQuestInterface", "CONFIG")) {
       my $cqClass = "CQWebJava";
       if (&GetParm("CQCC_CQWEB_VERSION") eq "1.0") {
           $cqClass = "CQWeb";
       }
       if ($perl5006 && (&GetParm("CQCC_CQWEB_ONLY") ne "TRUE")) {
           for (@INC) {
              if (-e $_ . "/" . "CQPerlExt.pm") {
                  $cqClass = "CQPerlAPI";
                  last;
              }
          }           
       }
       &SetConfigParm("ClearQuestInterface", "CQCC::$cqClass");
    } 

    # Would be improved if it just looked for the language, English
    # and then made the class name from that. For now assume that the
    # default of MsgCat_English is fine. Can still be overridden
    # in configuration script.

    # &SetConfigParm("MessageCatalogInterface",
    #     "CQCC::MsgCat_English");

    IODebugPrint("BASIC", "TriggerCQCC::ConfigureClasses: \n" .
        "   Running Perl $] \n " .
        "   Trigger: " . &GetParm("TriggerInterface") . "\n" .
        "   ClearCase: " . &GetParm("ClearCaseInterface") . "\n" .
        "   ClearQuest: " . &GetParm("ClearQuestInterface") . "\n" .
        "   MessageCatalog: " . &GetParm("MessageCatalogInterface") .
              "\n" );
    return;
}

######################################################################
#::method CheckLocalQuery

=over 6

=item C<ConfigureClasses()>

This method checks the local query definition to make sure the first
field is ID. This method should only be called once at very beginning
time.

=back

=cut

######################################################################

sub CheckLocalQuery {

    my $self = shift;

    if( $self->{CHECK_LOCAL_QUERY} == 1) {
        return;
    }
    
    $self->{CHECK_LOCAL_QUERY} = 1;          

    my $schema = 
        &GetCQSchemaByEntity(&GetParm("CQCC_ENTITY"));
    my @selectQueryList = $schema->GetQueryList();   

    for (@selectQueryList) {
        my ($resultsSet, $resultsFormat);

        # Get the results set description
        ($resultsSet, $resultsFormat) = 
                $schema->GetResultSet($_);

        # Need to make sure the first field is ID in query
        # The ID field name itself can't be changed because 
        # it is provided by the CQ packages. 
        my ($firstField) = split(",",$resultsSet);
        if (uc($firstField) ne "ID" && uc($firstField) ne "<ID>") {
            &IOMessage("ERROR", ["TCQCC_E_ID_MUST_BE_FIRST", $_]);
            # Don't return - complain about ALL problems
        }
    }
}

######################################################################
#::method Action_Cancel

=over 6

=item C<Action_Cancel()>

This method cancels the trigger operation and exits with a return
code of 1.

=back

=cut

######################################################################

sub Action_Cancel {
    my $self = shift;

    # Canceling an individual operation should not require any caching
    # for next files in series (by definition) nor would we expect to 
    # see the post-checkin trigger since we are canceling the operation
    &ForceParm("SAVE_SERIES_CACHE","FALSE");
    $self->SaveCache();
    &IOExitHandler(1);
}

######################################################################
#::method Action_Cancel_ApplyToAll

=over 6

=item C<Action_Cancel_ApplyToAll()>

This method cancels the trigger operation and exits with a return
code of 1. If an operation required any cleanup (i.e. canceling
a checkout will require an uncheckout), IOExitHandler will call the
cleanup routine. 

=back

=cut

######################################################################

sub Action_Cancel_ApplyToAll {
    my $self = shift;

    # Set the flag to apply to remaining items in the batch
    # clear if there is nothing else comint.

    if (&GetParm("CLEARCASE_END_SERIES") == 0) {
       &ForceParm("CACHE_APPLY_TO_ALL", 3);
    } else {
       &ForceParm("SAVE_SERIES_CACHE","FALSE");
    }
    &IODebugPrint("BASIC", "TriggerCQCC::Action_Cancel_ApplyToAll: "
       . "Setting ApplyToAll for next batch to " .
       &GetParm("CACHE_APPLY_TO_ALL") .
       "\n    (0 means end, non-zero means more to come)" );

    $self->SaveCache();
    &IOExitHandler(1);
}

######################################################################
#::method Action_Commit

=over 6

=item C<Action_Commit()>

This method commits association changes to the ClearCase and
ClearQuest databases then exits. If associations are required but not
provided it displays an error message rather than exiting.

=back

=cut

######################################################################

sub Action_Commit {
    my $self = shift;
    if ($self->CheckRestrictions()) {
       return $self->Commit_Assocs();
    }
    return;
}

######################################################################
#::method CheckRestrictions

=over 6

=item C<CheckRestrictions()>

This method performs a restrictions check on the desired associations
to make sure that at least some associations are provided if required
and that the associations pass the Restrictions and mastership queries. 

Returns 1 if the check succeeded, 0 if it failed. 

=back

=cut

######################################################################

sub CheckRestrictions {
    my $self = shift;
    my @desiredAssocList = $self->{DESIRED_ASSOCS}->GetAssocs();

    #If Required, then error if no desired associations
    if ($#desiredAssocList == -1 && &GetParm("REQUIRED") eq "TRUE") {
        &IOMessage("WARN", ["TCQCC_E_ASSOC_NOT_SPECIFIED"]);
        return 0;
    }

    # Check for multiple associations being permitted
    if ($#desiredAssocList > 0 && 
            &GetParm("CQCC_MULTIPLE_ASSOCS") eq "FALSE") {
        &IOMessage("WARN",["TCQCC_E_NO_MULTI_ASSOCS"]);
        return 0;
    }

    # Restrictions Query (if enabled)
    if ($#desiredAssocList > -1) {
        if (! $self->{CQ}->CheckRestrictions($self->{DESIRED_ASSOCS})) {
            return 0;
        }
        # Cache the most recent successful restrictions test in case
        # later files in the batch can be expedited. 

        &ForceParm("CACHE_RECENT_ASSOCS", 
            join(",", time(), $self->{DESIRED_ASSOCS}->GetAssocs()));
    }
    return 1;
}

######################################################################
#::method CheckRestrictionsCache

=over 6

=item C<CheckRestrictionsCache()>

This internal method performs a restrictions check on the desired
associations by comparing the currently list to the most recent list
of associations that passed restrictions within the last few
minutes. It will only work with CQCC_AUTO_ASSOCIATE or
CQCC_COMMENT_PATTERN based association selection.

Returns 1 if the check succeeded, 0 if it failed. 

=back

=cut

######################################################################

sub CheckRestrictionsCache {
    my $self = shift;
    my $cache = &GetParm("CACHE_RECENT_ASSOCS");
    my $timeout = &GetParm("CQCC_RESTRICTIONS_TIMEOUT");
    my $setAssocs = 0;
    my $result = 0;

    # Fail if recent associations weren't cached or timeout is disabled
    if (! defined $cache || $timeout <= 0) {
        return 0;
    } 

    ########################################
    # Process Environment variables: 

    if ($self->ProcessAutoAssocEVs(0) eq 0) {
        return 0;
    } 

    ########################################
    # If timestamp is within the timeout range
    # and the desired assocs are all on the recent list
    # then return ok 

    my ($timestamp, @recentAssocs) = split(/,/, $cache);
    if ((time() - $timestamp) < $timeout) {
        my ($newAssoc, $recentAssoc);
        my @desiredAssocList = $self->{DESIRED_ASSOCS}->GetAssocs();
        $result = 1;
        for $newAssoc (@desiredAssocList) {
            my $found = 0;
            for $recentAssoc (@recentAssocs) {
                if ($newAssoc eq $recentAssoc) {
                    $found = 1;
                    last;
                }
            }
            if ($found eq 0) {
                $result = 0;
                last;
            }
        }
    }
    return $result;
}

######################################################################
#::method CleanCacheDir

=over 6

=item C<ClearCacheDir(directory, maxAgeHours, cleanAgeHours)>

This internal method cleans up obsolete files in a given directory
that are older than maxAgeHours. When the operation is it will 
use an operation lock that leaves behind an empty lock file in the 
directory. 

If cleanAgeHours is non-zero, the operation will only run if the operation
lock doesn't exist or is at least cleanAgeHours old.  This way the
operation will run no more than once every cleanAgeHours.

=back

=cut

######################################################################

sub CleanCacheDir {
    my ($self, $dir, $maxAgeHours, $cleanAgeHours) = @_;
    my $lockFile =  $dir . "/.cqcc_clean";
    my $timestamp = $lockFile . ".time";
    my $cleanAgeSecs = $cleanAgeHours * 60 * 60;
    my $maxAgeSecs = $maxAgeHours * 60 * 60;
    my @stampInfo = stat($timestamp);
    my $stampModifiedDate = $stampInfo[9];

    if ($cleanAgeHours) {
        if (-e $timestamp && ((time()-$stampModifiedDate) < $cleanAgeSecs)) {
            return;
        }
    }
    # If lock can be obtained, clean out old files. 
    # Autobreak existing lock if more than specified age

    if (&IOLockOperation($lockFile, $IO_LOCKOP_NET, 
           "AUTOBREAK_MINS" => &GetParm("CM_LOCK_BREAK_MINS"))) {
        opendir CACHEDIR, $dir;
        my @allFiles = readdir CACHEDIR;
        closedir CACHEDIR;
        my $cleanCount;
        for (@allFiles) { 
            my $file = $_;
            # Clean up old series cache files (cqcc_cache.*) and
            # cqsession files (.cqsession.*) that are older than X days

            if ($file =~ /cqcc_cache\.(\S*)/ ||
                $file =~ /ccase_cache\.(\S*)/ ||
                $file =~ /cqsession\.(\S*)/) {
                my $path = $dir . "/" . $file;
                my @fileInfo = stat($path);
                if ((time() - $fileInfo[9]) > $maxAgeSecs) {
                    &IODebugPrint("DETAILED",
                        "CleanCache: Deleting obsolete cache $path " . 
                        " age exceeds $maxAgeHours hours");
                    unlink($path);
                    $cleanCount++;
                }
            }
        }
        &IOFileWrite($timestamp, "Removed $cleanCount files\n");
        &IOLockOperation($lockFile, $IO_LOCKOP_UNLOCK);
    }
}

######################################################################
#::method Commit_Assocs

=over 6

=item C<Commit_Assocs()>

This internal method is used by Action_Commit and Op_uncheckout to make
CC hyperlinks and CQ associations, either immediate or deferred.

=back

=cut

######################################################################

sub Commit_Assocs {
    my $self = shift;
    my @desiredAssocList = $self->{DESIRED_ASSOCS}->GetAssocs();
    my ($addAssocs,$delAssocs) = 
        $self->{EXISTING_ASSOCS}->CreateEditLists($self->{DESIRED_ASSOCS});
    &ForceParm("CACHE_DESIRED_ASSOCS", join(",", @desiredAssocList));

    # ClearQuest associations with ClearCase versions change
    # between checkout and checkin so we need to delete the existing
    # associations and remake them afresh.
    # ClearCase links to ClearQuest defects can be updated more
    # efficiently by just updating the delta between checkout and checkin

    # PostCheckin_Commit: Delay changes to CC and CQ until PostCheckin
    # trigger is called. 
    # If CQCC_POSTCHECKIN_COMMIT enabled, don't make changes now,
    # but save cache for later use. 

    if ((&GetParm("CQCC_POSTCHECKIN_COMMIT") eq "TRUE" ||
         &GetParm("CQCC_POSTCHECKIN_VERIFY") eq "TRUE") &&
        $self->{OPERATION} eq "checkin" && $self->{POSTOP} eq 0) {
        $self->SaveCache();
        &IOExitHandler(0);
    } 

    # If CC changes succeed then record CQ changes, else return with error
    # and seek more user interaction, including cancelation.

    my $ccResult = $self->{CC}->EditLinks($addAssocs, $delAssocs);
    if ($self->Commit_CQAssocs() && $ccResult) {
        $self->SaveCache();
        $self->Commit_TryAutoRun();
        &IOExitHandler(0);
    }
    # Return because commit operation failed but might be rectified
    # by user interaction when the GUI is active. 
    return 0; 
}

######################################################################
#::method Commit_CQAssocs

=over 6

=item C<Commit_CQAssocs()>

Commit associations to the ClearQuest database immediately or log
the information to the log file for delayed posting later. 

=back

=cut

#####################################################################

sub Commit_CQAssocs {
    my ($self) = @_;

    # STORE CQ ASSOCIATIONS: 
    # 
    # If batch file option enabled then write the CQ associations out as
    # batch entries for later processing

    my $result = 0;
    my $batchFileName = $self->{BATCHFILE};
    my ($addAssocs,$delAssocs);
    if ($self->{OPERATION} eq "edit") {
        ($addAssocs,$delAssocs) = 
            $self->{EXISTING_ASSOCS}->CreateEditLists($self->{DESIRED_ASSOCS});
    } else {
        $delAssocs = $self->{EXISTING_ASSOCS};
        $addAssocs = $self->{DESIRED_ASSOCS};
    }
    # Check numbers of adds or deletes before getting batch lock
    my @addList = $addAssocs->GetAssocs();
    my @delList = $delAssocs->GetAssocs();

    # Validate the pattern of path, vobid and oid
    my $vol = $self->{VO_LIST};
    my ($path,$vobid,$oid,$oldid) = split(",",$vol);
    my $validation_result = $self->{CC}->ValidateVOList("Catched VO", $path, $vobid, $oid); 

    if ($batchFileName ne "" && (($#addList > -1) || ($#delList > -1))) {
        my ($op, $cqEntry);
        # Operation lock will lock batchFileName.LOCK file which should
        # avoid need to get write lock on batchFileName itself
        &IOLockOperation($batchFileName, $IO_LOCKOP_NET,
            "AUTOBREAK_MINS" => &GetParm("CM_LOCK_BREAK_MINS"),
            "WAIT_WARN_MINS" => &GetParm("CM_LOCK_WARNING_MINS"));
        if (! open (OUTFILE, ">>$batchFileName")) {
            # Can't write to batch file so report error and 
            # then proceed to next section to immediately make the changes
            
            &IOMessage("ERROR",["TCQCC_E_BATCH_OUTPUT_FAILED", 
                       $batchFileName]);
        } else {
            &IODebugPrint("BASIC", "TriggerCQCC::Commit_CQAssocs (" . 
                          $self->{OPERATION} . ") saving CQ " . 
                          "association info to $batchFileName");
            for $op ($addOp, $delOp) {
                my $assocList = ($op eq $addOp) ? $addAssocs : $delAssocs;
                my @entityDBlist = $assocList->GetAssocs();
                for $cqEntry (@entityDBlist) { 
                    if ($op eq $delOp && ($self->{OPERATION} ne "edit")) {
                        my ($basePath,$suffix) = 
                            split($self->{DELIMITER},$path);
                        $vol = join(",", $basePath, $vobid, $oid, $oldid);
                    }
                    print OUTFILE "<CQ>$cqEntry $op " . $vol . "\n";
                }
            }
            close OUTFILE;
            $result = 1;
        }
        &IOLockOperation($batchFileName, $IO_LOCKOP_UNLOCK);
    }
    # Post immediately if batch processing not enabled or failed
    if ($result eq 0) {
       # Uncheckout won't connect unless this is needed
       $self->Connect("ClearQuest");
       $result = $self->{"CQ"}->EditAssocs($self->{VO_LIST}, 
                                           $addAssocs, $delAssocs);
    }
    return $result;
}

######################################################################
#::method Commit_TryAutoRun

=over 6

=item C<Commit_TryAutoRun()>

This methods checks to see if the conditions are right to automatically
run the batch job to commit previously logged transactions, usually at 
the end of a series. 

When appropriate do an autorun: if no batch file is available, don't 
bother and exit without error; otherwise call Op_batch which exits as
appropriate. 

If not appropriate for an autorun, return. 

=back

=cut

######################################################################

sub Commit_TryAutoRun {
    my ($self) = @_;
    my $assocFile = $self->{BATCHFILE};

    # IF these conditions are all true then run the batch job now: 
    # 
    # IF AssociationFile is in use(batch-mode enabled)
    # AND trigger operation allows autorun (not checkin-preop)
    #     OR Pre-checkin trigger, not PostCI-Commit or PostCI-Verify
    #        and already connected to CQ 
    #        then its a Single-Checkin operation, so save extra CQ login
    # AND autorun is enabled
    # AND user series flag is not suppressing series-end handling
    # AND (last entry in the series OR Autoflushing is enabled)

    my $checkinPreOp = ($self->{OPERATION} eq "checkin" && 
                        $self->{POSTOP} eq 0) ? 1 : 0;
    my $endSeries = &GetParm("CLEARCASE_END_SERIES");
    my $insideSeries = (defined $endSeries && $endSeries eq 0) ? 1 : 0;
    if (($assocFile ne "") &&
        ((!$checkinPreOp) ||
         ($checkinPreOp && 
          &GetParm("CQCC_POSTCHECKIN_COMMIT") eq "FALSE" &&
          &GetParm("CQCC_POSTCHECKIN_VERIFY") eq "FALSE" &&
          defined $self->{"CQ"})
        ) &&
        (&GetParm("CQCC_ASSOC_BATCH_AUTORUN") eq "TRUE") &&
        (&GetParm("CQCC_ASSOC_BATCH_SERIES") ne "TRUE") &&
        (($insideSeries eq 0) || 
         (&GetParm("CQCC_ASSOC_BATCH_ENABLE") eq "AUTOFLUSH")))
    {

        &IODebugPrint("BASIC", "TriggerCQCC::" . $self->{OPERATION} . 
            "  Automatically run batch job from " . $assocFile);

        # Op_batch will report an error if the file doesn't exist so
        # confirm that first. 
        # If non-existent file then no work to do so exit with no error. 
        if (! -e $assocFile ) {
            $self->SaveCache();
            IOExitHandler(0);
        } else {
            #Op_batch is not expected to return
            $self->{AUTORUN}=1;
            $self->Op_batch("do", $assocFile);
        }
    };
    return 0;
}

######################################################################
#::method Commit_TryNoCQLogin

=over 6

=item C<Commit_TryNoCQLogin()>

This method provides a common operation handler for checkin and checkout
when the operation can be performed without user interaction or CQ 
communications because the user has selected ApplyToAll for the same
set of associations and associations are being recorded to a batch file
for later application. It expects that ClearCase has been connected and 
that the ExistingAssociations information has been set already. 

If commit operation is appropriate, then Commit_Assocs is attempted;
if that returns, then exit with an error. 

Otherwise return and let the normal operation run instead. 

=back

=cut

######################################################################

sub Commit_TryNoCQLogin {
    my $self = shift;
    my $assocFile = $self->{BATCHFILE};

    if ($assocFile eq "") {
        return 0;
    }

    # If batch processing is enabled and we are in the middle of a series
    # (ClearCase or user defined), then skip connecting to CQ and write
    # associations to the log file. This skips the GUI setup, requirements
    # check and other initialization not needed when the same associations
    # are being applied to a series of checkouts or checkins. 

    my $series = &GetParm("CLEARCASE_SERIES_ID");

    ##################################################
    # IF trigger operation can use this option 
    # AND cached series is still running so we are inside a series
    #     and we don't need to keep checking the same CQ associations
    #     AND user selected ApplyToAll using the same set of associations
    # -or- CQCC_AutoAssociate or CommentPattern are enabled and 
    #      the desired associations all recently passed restrictions test
    ##################################################

    my $inBatchApplyToAll = 
        (defined $series && $series ne "" &&
            &GetParm("CACHE_SERIES") eq $series &&
            (&GetParm("CACHE_APPLY_TO_ALL") == 1)) ? 1 : 0;

    if ($inBatchApplyToAll || $self->CheckRestrictionsCache()) {
       &IODebugPrint("BASIC",
           "TriggerCQCC::Op_" . $self->{OPERATION} . 
           $self->{POSTOP} ? "(POST)" : "(PRE)" .
           " Expedite operation when using " . $assocFile);

        # Commit_Assocs will connect to ClearQuest if needed

        $self->Connect("ClearCase");
        $self->{VO_LIST} = $self->{CC}->GetVOList(
            $self->{OPERATION} eq "checkin" ? "XPN" : "PN");
        &ForceParm("CACHE_VOLIST", $self->{VO_LIST});
        if ($self->{OPERATION} eq "checkin") {
            $self->{EXISTING_ASSOCS} = $self->{CC}->ExistingAssocList();
        }
        $self->{DESIRED_ASSOCS}->SetAssocs(
                 split(",", &GetParm("CACHE_DESIRED_ASSOCS")));
        &ForceParm("DesiredAssocList",
            join(", ", $self->{DESIRED_ASSOCS}->GetBugList()));

        # Commit_Assocs is not expected to return
        $self->Commit_Assocs();   # Skip restriction checks

        # If return from Commit_Assocs operation failed so return error
        # This operation is only valid when processing a series of 
        # records non-interactively. 

        IOExitHandler(1); 
    }
    return 0;
}

######################################################################
#::method Action_ClearCQLogin

=over 6

=item C<Action_ClearCQLogin()>

This action prompts the user for confirmation then clears the existing
CQ login information to force a new login. It will automatically
remove the cache file and attempt to reconnect and reprompt for new
information.

=back

=cut

######################################################################

sub Action_ClearCQLogin {
    my $self = shift;

    my ($abort,$answer) = &IOPromptYesNo(["CQ_P_LOGIN_CLEAR"]);
    if (! $abort && $answer eq 1) {
        $self->{CQ}->ClearLogonParms();
        $self->{CQ} = undef;
        $self->Connect("ClearQuest");
        if (defined $self->{GUI}) { 
            $self->GUI_Action_Refresh();
        }
    }
}

######################################################################
#::method Action_Commit_ApplyToAll

=over 6

=item C<Action_Commit_ApplyToAll({UseCheckouts})>

This method sets the cache flags to apply the same values to the
remaining versions in the same batch then calls the commit action.

The optional UseCheckouts argument is used to change how ApplyToAll
will work with files in the batch. By default it will override
the current associations (assumed 0 if undefined); if non-zero it
will use the existing checked out associations.

=back

=cut

######################################################################

sub Action_Commit_ApplyToAll {
    my ($self, $useCheckouts) = @_;

    # Set the flag to apply to remaining items in the batch
    # clear if there is nothing else comint.

    if (&GetParm("CLEARCASE_END_SERIES") eq 0) {
       &ForceParm("CACHE_APPLY_TO_ALL",
           ((defined $useCheckouts && $useCheckouts > 0) ? 2 : 1));
    }
    else {
       &ForceParm("CACHE_APPLY_TO_ALL", 0);
    }

    &IODebugPrint("BASIC", "TriggerCQCC::Action_Commit_ApplyToAll: "
       . "Setting ApplyToAll for next batch to " .
       &GetParm("CACHE_APPLY_TO_ALL") .
       "\n    (0 means end, non-zero means more to come)" );

    $self->Action_Commit();

    # If Action_Commit succeeds, it will exit from the trigger
    # Otherwise it reported an error so clear the flag until
    # the user sets it again.

    &ForceParm("CACHE_APPLY_TO_ALL",0);
}

######################################################################
#::method Action_Commit_ApplyToAll_UseCheckouts

=over 6

=item C<Action_Commit_ApplyToAll_UseCheckouts()>

This method is a variation on Action_Commit_ApplyToAll which
asks that the remaining items in the batch use their original
checkout associations rather being overridden by a new set of
associations.  It calls Action_Commit_ApplyToAll with a flag to
cause it to use the checkout values (1). 

=back

=cut

######################################################################

sub Action_Commit_ApplyToAll_UseCheckouts {
    my $self = shift;
    $self->Action_Commit_ApplyToAll(1);
}

######################################################################
#::method Action_Help

=over 6

=item C<Action_Help(topic)>

This method presents help on the given topic then returns.

=back

=cut

######################################################################

sub Action_Help {
    my ($self, $topic) = @_;

    if (! defined $topic || $topic eq "") {
        $topic = $self->{HELP_CONTEXT};
    }
    &IOMessage("INFO", [ $topic ]);
    return;
}

######################################################################
#::method Action_Noop

=over 6

=item C<Action_Noop>

This method presents a dividing line in the menus and otherwise
is inactive.

=back

=cut

######################################################################

sub Action_Noop {
    return;
}

######################################################################
#::method Dialog_Cancel

=over 6

=item C<Dialog_Cancel()>

This method cancels the trigger operation and exits with a return
code of 1. If a batch is being processed, it will offer the choice
to cancel the remaining items in the batch without further 
interaction.

=back

=cut 

######################################################################

sub Dialog_Cancel {
    my $self = shift;
    my $choice = "Action_Cancel";
    my (@args, $abort);

    #<if batch mode then menu of choices>

    my $series = &GetParm("CLEARCASE_SERIES_ID");
    my $endSeries = &GetParm("CLEARCASE_END_SERIES");
    if ((defined $series && $series ne "" && $series ne "0") &&
        (defined $endSeries && $endSeries ne "1")) {

        my $exit;    
        until ($exit) {  
        my ($abort, $choiceTag) = 
           &IOListSelect(0, [ "Dialog_Cancel_Menu" ],
           [ "Action_Return" ],
           [ "Action_Cancel" ],
           [ "Action_Cancel_ApplyToAll"],
           # [ "Action_Help", "Dialog_Cancel_Help" ]
           );

        if ($abort) { 
            ($choice, @args) = ("Action_Return");
                $exit = 1;
            } else {
                if ($choiceTag) {
                    ($choice, @args) = @{ $choiceTag };
                    $exit = 1;
                } else {
                    &IOMessage("WARN", ["TCQCC_E_SELECTION"]);
                }
            }
        }
    }

    # If not action_return, then call choice method
    if ($choice ne "Action_Return") {
        my $cmd = "\$self->$choice(" . join(",",@args) . ")";
        my $result = eval $cmd;
    }
}

######################################################################
#::method Dialog_Commit

=over 6

=item C<Dialog_Commit()>

This method processes the choice to commit, making final tests on the
appropriateness of the associations being made, etc. If the current
version is part of a batch, additional choices are offered to
a) apply the associations to the current version only, b) apply the
same associations to the rest of the batch or c) process the rest
of the batch using the original checked out associations.

=back

=cut

######################################################################

sub Dialog_Commit {
    my $self = shift;
    my $choice = "Action_Commit";
    my @args;

    #<if batch mode then menu of choices>
    my $series = &GetParm("CLEARCASE_SERIES_ID");
    my $endSeries = &GetParm("CLEARCASE_END_SERIES");
    if ((defined $series && $series ne "" && $series ne "0") &&
        (defined $endSeries && $endSeries ne "1")) {
        my @choiceTagList =
           ( [ "Action_Return" ],
             [ "Action_Commit" ],
             [ "Action_Commit_ApplyToAll"]);
        my $existing = $self->{EXISTING_ASSOCS}->GetBugList();
        my $desired =  $self->{DESIRED_ASSOCS}->GetBugList();

        if ($self->{OPERATION} eq "checkin" && 
                $existing eq $desired) {
            push @choiceTagList,
               [ "Action_Commit_ApplyToAll_UseCheckouts" ];
        }
#       push @choiceTagList,[ "Action_Help", "Dialog_Commit_Help" ];

        my ($abort, $choiceTag) = 
            &IOListSelect(0, [ "Dialog_Commit_Menu" ],@choiceTagList);
        if ($abort) { 
            ($choice, @args) = ("Action_Return");
        } else {
            ($choice,@args) = @{ $choiceTag };
        }
    }

    # If not action_return, then call choice method
    if ($choice ne "Action_Return") {
        # Note: None of the current actions being called require
        # an argument and the OPERATION argument on _ApplyToAll
        # doesn't work if it isn't quoted. 
        my $cmd = "\$self->$choice()";
        my $result = eval $cmd;
    }
    return;
}


######################################################################
#::method Dialog_EditDesiredAssocs

=over 6

=item C<Dialog_EditDesiredAssocs()>

This method allows the user to delete individual items from the
desired associations list and update it. Comparable to the
AssociationDialog "AssociatedRecords" and "Disassociate" button
provided by the V1 integration on Windows. 

=back

=cut

######################################################################

sub Dialog_EditDesiredAssocs {
    my $self = shift;

    my @desiredAssocList = $self->{DESIRED_ASSOCS}->GetBugList();

    if ($#desiredAssocList < 0) {
        &IOMessage("WARN", ["TCQCC_E_NO_ASSOCS"]);
        return;
    }

    my @CRlist =
       $self->{CQ}->PerformQuery(&GetParm("CQCC_DATABASE"),
         &GetParm("CQCC_ENTITY"), undef,
         [ "<ID> in " . join(",", @desiredAssocList) ],
         "STANDARD", "FALSE");

    # Add any items NOT queried but on desired list (other db's,etc)
 
    my $desiredAssoc;
    my @addList; 

    if ($#desiredAssocList != $#CRlist) { 
        for $desiredAssoc (@desiredAssocList) { 
            my $found = 0;
            for (@CRlist) { 
                my ($id) = split(" ",$_);
                if ($desiredAssoc eq $id) { 
                    $found = 1;
                }
            }
            if (! $found) { 
                push @CRlist, $desiredAssoc . " ";
            }
        }
    }

    my ($abort, @choices) =
       &IOListSelect(1, ["TCQCC_P_QUERY_ASSOCLIST", "PARM:NO_TEXT"], @CRlist);

    if (! $abort && @choices) {
       my %deleteList = ();
       for (@choices) {
           # Pattern: <name> <whitespace>, $1 is <name>
           if ($_ =~ m/^(\S+)\s/) {
               $deleteList{$1} = 1;
           }
       }
       my @newList;
       for (@desiredAssocList) {
           if (! $deleteList{$_}) {
               push @newList, $_;
           }
       }
       $self->{DESIRED_ASSOCS}->ProcessBugString(
           &GetParm("CQCC_ENTITY"), join(" ", @newList));
       &ForceParm("DesiredAssocList",
           join(", ", $self->{DESIRED_ASSOCS}->GetBugList()));
    }
}

######################################################################
#::method Dialog_SelectByTypein

=over 6

=item C<Dialog_SelectByTypein()>

This method implements a simple type-in interface to select bugs,
performing parsing to get the needed lists, and offering the user
the choice to change mode to other alternatives like choosing from
a list menu.

=back

=cut

######################################################################

sub Dialog_SelectByTypein {
    my $self = shift;

    my $cqname = &GetParm("CLEARCASE_PN");
    my $done = 0;
    my ($abort, $a, $r);
    my $opt_or_reqd;

    while (! $done) {

        $opt_or_reqd = (&GetParm("REQUIRED") eq "TRUE") ?
                           "TCQCC_P_ENTER_BUGS_NOWARN_REQD" : "TCQCC_P_ENTER_BUGS_NOWARN_OPT";

        ($abort,$a) = 
            &IOPrompt([$opt_or_reqd, $self->{OPERATION}, $cqname]);

        $done = 1;
        if (! $abort) {
            if (length($a)) {
                if ($a eq "?") {
                    $self->Action_Help("TCQCC_I_GETBUGS_HELP");
                    $done = 0;

                } else {
                    my @tokens = &IOParseListString($a);
                    if ($#tokens > 0 && 
                        &GetParm("CQCC_MULTIPLE_ASSOCS") eq "FALSE") {
                        &IOMessage("WARN",
                            ["TCQCC_E_NO_MULTI_ASSOCS"]);
                        $done = 0;
                    }
                    $r = $self->{DESIRED_ASSOCS}->
                        ProcessBugString(
                           &GetParm("CQCC_ENTITY"),$a);
                    &ForceParm("DesiredAssocList",
                        join(", ",
                            $self->{DESIRED_ASSOCS}->GetBugList()));

                    # Check for bad grammar
                    if (!$r) {
                        &IOMessage("INFO", ["TCQCC_E_BUGLIST_INPUT", $a,
                                           &GetParm("CQCC_DATABASE"),
                                           &GetParm("CQCC_ENTITY") ]);
                        $done = 0;
                    }
                }
            }
        }
    }
    return ($abort);
}

######################################################################
#::method Dialog_SelectByQuery

=over 6

=item C<Dialog_SelectByQuery()>

This method implements a query display with a simple type-in interface
to select bugs, performing parsing to get the needed lists, and
offering the user the choice to change mode to other alternatives like
choosing from a list menu.

=back

=cut


######################################################################

sub Dialog_SelectByQuery {
    my $self = shift;

    my $done = 0;
    my ($result, @choices, @numbers);

    my $queryType = &GetParm("CQCC_QUERY_TYPE");
    my $retType = $queryType eq "LOCAL" ? "STRING" : "ARRAY";

    my @CRlist = 
       $self->{CQ}->PerformQuery(&GetParm("CQCC_DATABASE"),
         &GetParm("CQCC_ENTITY"), undef, 
         &GetParm("CQCC_QUERY_NAME"), &GetParm("CQCC_QUERY_NAME"),
         "FALSE", "RETURN" => $retType, 
         "QUERY_TYPE" => $queryType);

    # Need to determine formatting for ARRAY query types
    # and reprocess them into strings

    if (($#CRlist > 0) && $retType eq "ARRAY") {
        my @fieldLens = ();
        my ($field, $i, $recordFormat);

        shift @CRlist; # Skip first row with titles in it, not shown
        for (@CRlist) {
            my @fields = @{ $_ };
            for $i (0 .. $#fields) { 
               my $len = length($fields[$i]);
               if (! defined $fieldLens[ $i ] || 
                   $fieldLens[ $i ] < $len) {
                   $fieldLens[ $i ] = $len;
               }
            }
        }
        for (@fieldLens) { 
            my $len = $_ ? $_ : 0;
            $recordFormat .= "%-" . $len . "." . $len . "s ";
        }
        my @CRlistStrings;
        # Note: Assumes ID will be first field
        for (@CRlist) { 
           my @fields = @{ $_ };
           my $recordString = sprintf $recordFormat, @fields;
           push @CRlistStrings, $recordString;
        }
        @CRlist = @CRlistStrings;
    } 

    if ($#CRlist >= 0) {
       my $multiChoice = 0;
       if (&GetParm("CQCC_MULTIPLE_ASSOCS") eq "TRUE") {
          $multiChoice = 1;
       }
       my ($abort, @choices) = &IOListSelect($multiChoice,
               ["TCQCC_P_QUERY_RESULTS", "PARM:USER"], @CRlist);

       if ($abort) { 
          return;
       } 
       for (@choices) {
          $_ =~ m/(\S*) /;
          push @numbers, $1;
       }
       my $numberString = join(" ",@numbers);
       &IODebugPrint("BASIC", "TriggerCQCC::Dialog_SelectByQuery " .
           "Selected: $numberString\n");
       $self->{DESIRED_ASSOCS}->ProcessBugString(
           &GetParm("CQCC_ENTITY"), $numberString );
       &ForceParm("DesiredAssocList",
           join(", ", $self->{DESIRED_ASSOCS}->GetBugList()));
    } else {
       &IOMessage("WARN", ["TCQCC_E_QUERY_EMPTY",&GetParm("USER")]);
    }
    return;
}

######################################################################
#::method Dialog_SelectDatabase

=over 6

=item C<Dialog_SelectDatabase()>

This method allows the user to select which ClearQuest database
to connect to.

=back

=cut

######################################################################

sub Dialog_SelectDatabase {
    my $self = shift;

    my @dbList = &SupportedDatabaseList();
    # Older versions of perl don't handle Can() method
    # but it is more robust to use it when it is available. 
 
    if ($perl5006 && 
        ! @dbList && $self->{CQ}->can("GetDatabaseList")) {
       @dbList = $self->{CQ}->GetDatabaseList();
    }
    my ($abort, $choice);

    # Don't allow switching databases if there are existing bugs. 
    my @existingBugs; 
    if ($self->{EXISTING_ASSOCS}) { 
        @existingBugs = $self->{EXISTING_ASSOCS}->GetBugList();
    } 
    if ($#existingBugs > -1) {
        &IOMessage("WARN", ["TCQCC_E_EXISTING_BUGS_LOCK_DB"]);        
        return;
    } 
    
    if (@dbList) {
        ($abort, $choice) = &IOListSelect(0,
           ["Dialog_SelectDatabase", "PARM:CQCC_DATABASE"],
           @dbList);
    } else {
        my $abort;
        ($abort,$choice) =
            &IOPrompt(["Dialog_SelectDatabase","PARM:CQCC_DATABASE"]);
    }

    if (! $abort && $choice && $choice ne &GetParm("CQCC_DATABASE")) {
       # If changing database, then clear any chosen associations
       $self->{DESIRED_ASSOCS}->ProcessBugString("");
       &ForceParm("DesiredAssocList",
           join(", ", $self->{DESIRED_ASSOCS}->GetBugList()));
       # Would need to relogin to database
       $self->{CQ}->SetDatabase($choice);
       # If successful change, flag change as user initiated
       if (&GetParm("CQCC_DATABASE") eq $choice) {
           &ForceParm("CQCC_DATABASE", $choice, "USER_CHANGED" => 1);
       }
       my @entList =
          &SupportedEntityList(&GetParm("CQCC_DATABASE"));
       my $found;
       for (@entList) { 
           if ($_ eq &GetParm("CQCC_ENTITY")) {
               $found = 1;
           }
       }
       if (! $found) {
           &ForceParm("CQCC_ENTITY", $entList[0], "USER_CHANGED" => 1);
       }
       $self->QueryNameValidate("REFRESH");
       if ($self->{GUI}) {
           $self->GUI_UpdateQueryResults();
       }
    }
}

######################################################################
#::method Dialog_SelectEntity

=over 6

=item C<Dialog_SelectEntity()>

This method allows the user to select which ClearQuest entity
to make associations with, i.e. "defect".

=back

=cut

######################################################################

sub Dialog_SelectEntity {
    my $self = shift;

    my @entList =
        &SupportedEntityList(&GetParm("CQCC_DATABASE"));

    my ($abort,$choice);

    if (@entList) {
        ($abort, $choice) = &IOListSelect(0,
            ["Dialog_SelectEntity", "PARM:CQCC_ENTITY"], @entList);
    } else {
        ($abort, $choice)  = &IOPrompt(
            ["Dialog_SelectEntity", "PARM:CQCC_ENTITY"]);
    }

    if (! $abort && $choice) {
       &ForceParm("CQCC_ENTITY", $choice, "USER_CHANGED" => 1);
       # Force to clear the cached query name list and load it based on new entity
       $self->QueryNameValidate("REFRESH");
       if ($self->{GUI}) {
           $self->GUI_UpdateQueryResults();
       }
    }
}

######################################################################
#::method Dialog_SelectQueryName
######################################################################

sub Dialog_SelectQueryName {
    my ($self) = @_;    
    my $queryEnable = &GetParm("CQCC_QUERY_ENABLE");
    if (! &GetParm("CQCC_QUERY_ENABLE") eq "OFF") {
        return;
    }
    my @selectQueryList = $self->QueryNameList();

    # Present the list, get an answer, update the query if needed
    if ($#selectQueryList != -1) {
        my ($abort, $newQueryName) = 
          &IOListTreeSelect(0, "/", ["GUI_Title_SelectQuery"], 
                                @selectQueryList);
        if (! $abort && $newQueryName) {
            if ($newQueryName =~ /LOCAL\/(.*)/) { 
                &ForceParm("CQCC_QUERY_NAME", $1, "USER_CHANGED" => 1);
                &ForceParm("CQCC_QUERY_TYPE", "LOCAL", "USER_CHANGED" => 1);
            }
            else { 
                &ForceParm("CQCC_QUERY_NAME", $newQueryName, 
                           "USER_CHANGED" => 1);
                &ForceParm("CQCC_QUERY_TYPE", "CQ", "USER_CHANGED" => 1);
            }
            if ($self->{GUI}) {
                $self->GUI_UpdateQueryResults();
            }
        }
    }  
}

######################################################################
#::method Dialog_DisplayXPN

=over 6

=item C<Dialog_DisplayXPN()>

This method displays information about the currently selected
ClearCase version. Currently it doesn't display any additional
information but provides a way to show the current version 
being operated on, which is important when working on a batch
of changes. 

=back

=cut

######################################################################

sub Dialog_DisplayXPN {
    my $self = shift;

    &IOMessage("INFO", ["Dialog_DisplayXPN", "PARM:CLEARCASE_XPN"]);
}

######################################################################
#::method GUIBuild

=over 6

=item C<GUIBuild()>

This method is called as needed to initialize the Graphical User
Interface (GUI) for the integration using the IOPerlTK package. 

IOPerlTK provides the standard IOBase dialog methods (IOMessage, etc)
and a limited vocabulary of "canned" user interface components
consisting of panels (such as IOMainWindow or IOFrame), menus, and
manipulators (such as IOLabel, IOListbox, etc). IOPerlTK isolates the
Trigger from any direct dependency on the Perl/Tk widget package and
simplifies some aspects of creating a GUI for the needs of this
integration. See the IOPerlTK package for more information on how this
layer works.

The GUI presents a main association dialog panel consisting of the
following three frames: 

    Frame 1 - Association Records browser and command menu
    Frame 2 - Database and Record Type buttons and status display
    Frame 3 - Query results browser, query name, and association menu

This method may be modified or overridden by subclasses as necessary.
It is invoked by the GUIDisplay method which calls this method to
build the GUI then starts the main event loop.

=back

=cut

######################################################################

sub GUIBuild {
    my ($self) = @_;

    my $t1 = &TimerStart("TriggerCQCC::GUIBuild\n");
    my $IO = &IOstd(); # Get pointer to local IOPerlTK instance

    # Status shouldn't be problem since GetParm already did
    # conversion/ASCII check and operation is ASCII.

    my ($status,$ioTitle) = &IOConvertString(&GetParm("CLEARCASE_XPN"),
                                "CORE_CODEPAGE", "IO_CODEPAGE");
    
    $IO->Panel("Main", "", "IOMainWindow",
               GEOMETRY => "600x450",
               TITLE => $self->{OPERATION} . " " . $ioTitle);

    ########################################
    # Frame 1 - Association Records browser and command menu

    $IO->Panel("Frame1", "Main", "IOFrame",
               PACK=>[-side => "top",
                      -padx=>10, -pady=>10,
                      -fill => 'both' ]);
    $IO->Manipulator("AssocRecords", "Frame1", "IOListboxGrid",
               TITLE=>["GUI_Title_AssocRecords"], 
               MODE =>"extended",
               HEIGHT => 7, 
               PACK =>[-side => "left", -anchor => "n", 
                      -expand => 1, -fill => 'both']);
    $IO->Menu("Menu1", "Frame1", "IOCommandMenu", 
               CALLBACK_OBJECT => $self,
               WIDTH => 24,
               MENULIST =>
                   # Button MsgTag     Callback
                   [ ["GUI_Commit"] => "Action_Commit",
                     ["GUI_Cancel"] => "Dialog_Cancel",
                     ["GUI_ApplyToAll"] => "GUI_Action_ApplyToAll",
                     ["GUI_Disassociate"]=>"GUI_Action_Disassociate",
                     ["GUI_ClearCQLogin"] => "Action_ClearCQLogin",
                     ["GUI_Help" ]=>[ "Action_Help", "GUI_HelpMain" ],
                   ],
               PACK=>[-side => "right", -anchor => 'n', -fill=>'x']);

    # Disable the ApplyToAll button if not needed
    my $series = &GetParm("CLEARCASE_SERIES_ID");
    my $endSeries = &GetParm("CLEARCASE_END_SERIES");
    if (! ((defined $series && $series ne "" && $series ne "0") &&
           (defined $endSeries && $endSeries ne "1"))) {
       $IO->MenuManageButtons("Menu1", "STATE", 
                              ["GUI_ApplyToAll"], "disabled");
    }

    ########################################
    # Frame 2 - Database and Record Type buttons and status display
    my @f2Pack = (-side => "left", -anchor => "n", -fill => 'y' );
    $IO->Panel("Frame2", "Main", "IOFrame",
               PACK=>[-side => "top", -fill => 'x', 
                      -padx => 10]);
    $IO->Manipulator("Database", "Frame2", "IOButton",
               TITLE=>["GUI_Database"],
               CALLBACK_OBJECT => $self,
               CALLBACK => "Dialog_SelectDatabase",
               PACK=>\@f2Pack);

    $IO->Manipulator("DatabaseName", "Frame2", "IOLabel",
               PARM=>"CQCC_DATABASE",
               TK => [ -background => 'white', 
                       -relief => 'sunken' ],
               PACK=>\@f2Pack);

    $IO->Manipulator("RecordType", "Frame2", "IOButton",
               TITLE=>["GUI_RecordType"],
               CALLBACK_OBJECT => $self,
               CALLBACK => "Dialog_SelectEntity",
               PACK=>\@f2Pack);

    $IO->Manipulator("RecordTypeName", "Frame2", "IOLabel",
               PARM=>"CQCC_ENTITY",
               TK => [ -background => 'white',
                       -relief => 'sunken' ],
               PACK=>\@f2Pack);

    ########################################
    # Frame 3 - Query results browser, query name,and association menu

    $IO->Panel("Frame3", "Main", "IOFrame",
               TK=>[-borderwidth => 2, -relief => 'groove'],
               PACK=>[-side => "bottom", -expand=>1, -fill=>'both',
                      -padx=>10, -pady=>10 ]);

    $IO->Panel("SubFrame3A", "Frame3", "IOFrame",
               PACK=>[-side => "top", -fill => 'x',
                      -pady => 10, -pady => 10]);

    $IO->Manipulator("QueryResultsLabel", "SubFrame3A", "IOLabel",
               VALUE=>["GUI_Title_Query"],
               PACK=>[-side => "left", -anchor => "n"]);

    $IO->Manipulator("QueryName", "SubFrame3A", "IOLabel",
               PARM=>"CQCC_QUERY_NAME",
               TK => [ -background => 'white', -relief => 'sunken' ],
               PACK=>[-side => "left", -anchor => "n", -fill => 'x']);

    $IO->Manipulator("QueryResults", "Frame3", "IOListboxGrid",
#               TITLE=>["GUI_Title_Results"],
               MODE => (&GetParm("CQCC_MULTIPLE_ASSOCS") eq "TRUE") ?
                    "extended" : "single",
               PACK=>[-side => "left", -anchor => "s",
                      -expand => 1, -fill => 'both']);
    $IO->Menu("Menu2", "Frame3", "IOCommandMenu", 
               CALLBACK_OBJECT => $self,
               WIDTH => 24,
                           # Button MsgTag     Callback
               MENULIST =>[["GUI_Browse"]=>"Dialog_SelectQueryName",
                           ["GUI_Associate"]=>"GUI_Action_Associate",
                           ["GUI_Refresh"]=>"GUI_Action_Refresh",
                           ["GUI_Typein"]=>"GUI_Action_Typein",
                          ],
               PACK=>[-side => "right", -anchor => "n", -fill=>'y']);

    if (&GetParm("CQCC_QUERY_ENABLE") eq "OFF") {
       $IO->MenuManageButtons("Menu2", "STATE",
                              ["GUI_Browse"], "disabled");
    } 

    ########################################
    # Logic callbacks, etc.

    $self->{GUI} = 1;
    $IO->SetIdleCallback($self, "GUI_Action_Refresh");
    &TimerStop($t1);
    return 0;
}

######################################################################
#::method GUI_Action_ApplyToAll

=over 6

=item C<GUI_Action_ApplyToAll()>

This method handles the call to apply the operation to the entire
current series. If the operation is a checkin and it is still valid 
to apply the check-out associations to all, the choice is offered to
the user. 

=back

=cut

######################################################################

sub GUI_Action_ApplyToAll {
    my ($self) = @_;
    my $choice = "Action_Commit_ApplyToAll";
    my (@args, $abort);

    if ($self->{OPERATION} eq "checkin") { 
        my $existing = $self->{EXISTING_ASSOCS}->GetBugList();
        my $desired =  $self->{DESIRED_ASSOCS}->GetBugList();
        if ($existing eq $desired) {
            my $exit;    
            until ($exit) {  
            my ($abort,$choiceTag) = 
                &IOListSelect(0, ["Dialog_Commit_Menu"],
                          ["Action_Commit_ApplyToAll"],
                          ["Action_Commit_ApplyToAll_UseCheckouts"]);
            if ($abort) {  
                return;
            }

                if ($choiceTag) {
                    ($choice, @args) = @{ $choiceTag };
                    $exit = 1;
                } else {
                    &IOMessage("WARN", ["TCQCC_E_SELECTION"]);
                }
            }
         }
     }

     my $cmd = "\$self->$choice(" . join(",",@args) . ")";
     my $result = eval $cmd;
     return;
}

######################################################################
#::method GUI_Action_Associate

=over 6

=item C<GUI_Action_Associate()>

This method copies the selected items on the Query results browser
to the desired associations list and updates the Associated records
browser to show the current list. 
If CQCC_MULTIPLE_ASSOC=FALSE, and there is already a CR in the 
designedAssociation list, return. 

=back

=cut

######################################################################

sub GUI_Action_Associate {
    my ($self) = @_;
    if (! defined $self->{GUI}) { 
        return;
    } 
    my $IO = &IOstd();
    my @desiredAssocList = $self->{DESIRED_ASSOCS}->GetAssocs();
    if (&GetParm("CQCC_MULTIPLE_ASSOCS") eq "FALSE") {
        if (scalar(@desiredAssocList) > 0) {
            &IOMessage("WARN",["TCQCC_E_NO_MULTI_ASSOCS"]);
            return ;
        }
    }

    my @addIDList;
    my @data = $IO->ManipManageValues("QueryResults", "GETVALUE");
    # NOTE: Assumes that "id" is the first entry in the query.
    for (@data) {
        my @fields = @{ $_ };
        my $id = GetParm("CQCC_ENTITY") . ":" . $fields[0];
        push @addIDList, $id;
    }
    # Ensure each item only appears once
    my %desiredIDs; 
    for (@desiredAssocList) { 
        $desiredIDs{$_} = 1;
    }
    for (@addIDList) { 
        $desiredIDs{$_} = 1;
    }
    $self->{DESIRED_ASSOCS}->SetAssocs(sort keys %desiredIDs);
    $self->GUI_UpdateAssocRecords();
}

######################################################################
#::method GUI_Action_Disassociate

=over 6

=item C<GUI_Action_Disassociate()>

This method deletes the selected items in the Associated records
browser from the desired associations list and updates the browser
appropriately. 

=back

=cut

######################################################################

sub GUI_Action_Disassociate {
    my ($self) = @_;
    if (! defined $self->{GUI}) { 
        return;
    } 

    my $IO = &IOstd();
    my @desiredAssocList = $self->{DESIRED_ASSOCS}->GetAssocs();
    my @delIDList;       

    my @data = $IO->ManipManageValues("AssocRecords", "GETVALUE");
    # NOTE: Assumes that "id" is the first entry in the query.
    for (@data) {
        my @fields = @{ $_ };
        my $id = GetParm("CQCC_ENTITY") . ":" . $fields[0];
        push @delIDList, $id;
    }
    my %desiredIDs; 
    for (@desiredAssocList) { 
        $desiredIDs{$_} = 1;
    }
    for (@delIDList) { 
        delete $desiredIDs{$_};
    }
    $self->{DESIRED_ASSOCS}->SetAssocs(sort keys %desiredIDs);
    $self->GUI_UpdateAssocRecords();
}

######################################################################
#::method GUI_Action_Refresh

=over 6

=item C<GUI_Action_Refresh()>

This method flushes the current query selection list for the Browse
operation and updates the two record browsers. 

=back

=cut

######################################################################

sub GUI_Action_Refresh {
    my ($self) = @_;
    if (! defined $self->{GUI}) { 
        return;
    } 

    &CheckLocalQuery();

    $self->{QUERY_NAME_LIST} = undef;
    $self->GUI_UpdateAssocRecords();
    $self->GUI_UpdateQueryResults();
}

######################################################################
#::method GUI_Action_Typein

=over 6

=item C<GUI_Action_Typein()>

This method calls the standard dialog to process user input to
type in a list of defect ids, then ADDS them to the desired 
association list rather than replacing the entire list. This is
done because the GUI provides a more reasonable way to interact
with the list to delete undesired associations. 

=back

=cut

######################################################################

sub GUI_Action_Typein {
    my ($self) = @_;    
    if (! defined $self->{GUI}) { 
        return;
    } 

    my @desiredAssocList = $self->{DESIRED_ASSOCS}->GetAssocs();
    $self->Dialog_SelectByTypein();
    my @addIDList = $self->{DESIRED_ASSOCS}->GetAssocs();

    # Ensure each item only appears once
    my %desiredIDs; 
    for (@desiredAssocList) { 
        $desiredIDs{$_} = 1;
    }
    for (@addIDList) { 
        $desiredIDs{$_} = 1;
    }
    $self->{DESIRED_ASSOCS}->SetAssocs(sort keys %desiredIDs);
    $self->GUI_UpdateAssocRecords();    
}

######################################################################
#::method GUI_UpdateAssocRecords

=over 6

=item C<GUI_UpdateAssocRecords()>

This method updates the Associated Records browser to reflect the 
current desired associations list. 

=back

=cut

######################################################################

sub GUI_UpdateAssocRecords {
    my ($self) = @_;
    if (! defined $self->{GUI}) { 
        return;
    } 

    my $IO = &IOstd();
    my @desiredAssocList = $self->{DESIRED_ASSOCS}->GetBugList();
    my (@CRlist, $desiredAssoc, @addList);

    if (@desiredAssocList && $#desiredAssocList > -1) { 
       @CRlist = $self->{CQ}->PerformQuery(
         &GetParm("CQCC_DATABASE"),
         &GetParm("CQCC_ENTITY"), undef,
         [ "<ID> in " . join(",", @desiredAssocList) ],
         "STANDARD", "FALSE", "RETURN" => "ARRAY");
    }

    # Add any items NOT queried but on desired list (other db's,etc)
    # Note: Similar to code in Dialog_EditDesiredAssocs. 
    # Have to subtract one because an query output array has extra row with
    # column headings. 

    if ($#desiredAssocList != ($#CRlist - 1)) { 
        for $desiredAssoc (@desiredAssocList) { 
            my $found = 0;
            for (@CRlist) { 
                my @fields = @{ $_ };
                # Note: Assumes ID is first field
                if (uc($desiredAssoc) eq uc($fields[0])) {
                   $found = 1;
                }
            }
            if (! $found) {
                # Note: Assumes ID is first field
                my @dummyRecord = ($desiredAssoc);
                push @CRlist, \@dummyRecord;
            }
        }
    }

    # Build up column listings for headings
    if ($#CRlist < 0) {
        my $schema = &GetCQSchemaByEntity(&GetParm("CQCC_ENTITY"));
        my ($resultsSet, $resultsFormat) = 
            $schema->GetResultSet("STANDARD");
        my @titles = split(",", $resultsSet);
        push @CRlist, \@titles;
    }
    $IO->ManipManageValues("AssocRecords", "SETRANGE", \@CRlist);
}                            

######################################################################
#::method GUI_UpdateQueryResults

=over 6

=item C<GUI_Action_UpdateQueryResults()>

This method updates the Query Results browser to reflect the 
latest query information returned by ClearQuest. 

=back

=cut

######################################################################

sub GUI_UpdateQueryResults {
    my ($self) = @_;
    if (! defined $self->{GUI}) { 
        return;
    } 

    my $IO = &IOstd();
    my @queryList = 
       $self->{CQ}->PerformQuery(&GetParm("CQCC_DATABASE"),
         &GetParm("CQCC_ENTITY"), undef, 
         &GetParm("CQCC_QUERY_NAME"), &GetParm("CQCC_QUERY_NAME"),
         "FALSE", "RETURN" => "ARRAY", 
         "QUERY_TYPE" => &GetParm("CQCC_QUERY_TYPE"));

    $IO->ManipManageValues("QueryResults", "SETRANGE", \@queryList);
}

######################################################################
#::method MenuBuild

=over 6

=item C<MenuBuild(menuName)>

A menu is a list consisting of a title message tag and one or more
choice message tags which can be translated through the message
catalog along with a set of arguments.  For example, 
  
   [Action_Help, "main_menu"] 

might be displayed as "3) Help for main_menu".  If the title or choice
message tags are NOT message tag arrays, then the literal value is
shown instead of being looked up through the message tag.

When the user makes a selection, the menu gets back the message tag.
The tag value is expected to be a trigger method name and will be
called with the arguments, i.e. $obj->Action_Help("main_menu").

This approach allows a subclass to redefine the menu by editing its
definition or to override any of its activities by overriding the
method that gets called back.

=back

=cut

######################################################################

sub MenuBuild {
    my ($self, $menuName) = @_;
    if (! defined $menuName) {
        $menuName = "MENU_ASSOCIATION";
    }

    my @assoc_menu = 
       ( ["TITLE_ASSOCIATION_MENU", "PARM:YES_TEXT"],
         ["Dialog_Commit"],
         ["Dialog_Cancel"],
         ["Action_Help", "" ],

         ["Action_Noop"],

         ["Dialog_SelectByQuery"],
         ["Dialog_SelectByTypein"],
         ["Dialog_EditDesiredAssocs", "PARM:DesiredAssocList" ],

         ["Action_Noop"],

         ["Dialog_SelectDatabase", "PARM:CQCC_DATABASE"],
         ["Dialog_SelectEntity", "PARM:CQCC_ENTITY"],
         ["Dialog_DisplayXPN", "PARM:CLEARCASE_XPN"],
         ["Action_ClearCQLogin"],

        );

    # Add QueryName option unless disabled
    if (&GetParm("CQCC_QUERY_ENABLE") ne "OFF") {
       push @assoc_menu,
           ["Dialog_SelectQueryName", "PARM:CQCC_QUERY_NAME"],
    }

    $self->{MENU_ASSOCIATION} = \@assoc_menu;

    return;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::TriggerCQCC()>

This method presets the objects attributes to read in the basic
command line arguments. The object is not active until the owner
calls its Initialize() method.

=back

=cut

######################################################################

sub new {
   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   my $self = new CQCC::TriggerBase();   #Make instance of superclass
   # Provides following instance variables of interest to trigger:
   #    CC = ClearCase interace
   #    CQ = ClearQuest interface

   # Attributes
   $self->{CLASSNAME} = "CQCC::TriggerCQCC";

   # Trim any enclosing quotation marks
   # NOTE: All references to CQCC_ASSOC_BATCHFILE must use this 
   # variable so that the same result is used throughout the script. 

   my $batchFileName = &GetParm("CQCC_ASSOC_BATCHFILE");
   if (&GetParm("CQCC_ASSOC_BATCH_ENABLE") ne "FALSE" &&
           $batchFileName ne "") {
       $batchFileName =~ s/\"//g;
       if (uc($batchFileName) eq "STANDARD") {
           $batchFileName = &IOPathname("PARM",".cqcc_assoc_batch");

           # If old user area has existing batch file copy it 
           # to central batch file to avoid loosing it and rename
           # it to .central to indicate what happened
           my $oldBatchFile = &IOPathname("PARM_USER",".cqcc_assoc_batch");
           if ((&GetParm("CQCC_CACHE_ROOT") ne "") && 
               ($oldBatchFile ne $batchFileName) &&
               (-e $oldBatchFile)) {
               &IOLockOperation($oldBatchFile, $IO_LOCKOP_NET,
                   "AUTOBREAK_MINS" => &GetParm("CM_LOCK_BREAK_MINS"),
                   "WAIT_WARN_MINS" => &GetParm("CM_LOCK_WARNING_MINS"));
               &IOLockOperation($batchFileName, $IO_LOCKOP_NET,
                   "AUTOBREAK_MINS" => &GetParm("CM_LOCK_BREAK_MINS"),
                   "WAIT_WARN_MINS" => &GetParm("CM_LOCK_WARNING_MINS"));
               if (-e $oldBatchFile) {
                   my @records = &IOFileRead($oldBatchFile);
                   &IOFileAppend($batchFileName,@records);
                   &IOFileOperation($IO_FILE_RENAME, "FATAL", 
                       "TriggerCQCC::new", 
                       $oldBatchFile, $oldBatchFile . ".central");
               }
               &IOLockOperation($batchFileName, $IO_LOCKOP_UNLOCK);
               &IOLockOperation($oldBatchFile, $IO_LOCKOP_UNLOCK);
           }
       }
       $self->{BATCHFILE} = $batchFileName;
       # Force PostCheckin-Commit if PostCheckin-Verify is used 
       # with batching
       if (&GetParm("CQCC_POSTCHECKIN_VERIFY") eq "TRUE") {
           &ForceParm("CQCC_POSTCHECKIN_COMMIT", "TRUE");
       }

   } else {
       $self->{BATCHFILE} = "";
   }

   my $delimiter = $ENV{"CLEARCASE_XN_SFX"};
   if (! defined $delimiter) {
       $delimiter = "@@";
   }
   $self->{DELIMITER} = $delimiter;

   # AssocList for desired result
   $self->{DESIRED_ASSOCS} = new CQCC::AssocList();

   # AssocList for existing associations
   $self->{EXISTING_ASSOCS} = new CQCC::AssocList();

   # Menu for association dialog
   $self->{MENU_ASSOCIATION} = undef;

   # List of permitted queries
   $self->{QUERY_NAME_LIST} = undef;

   # CC path/uuid/oid for current object
   $self->{VO_LIST} = undef;

   # Note if auto-run was triggered
   $self->{AUTORUN} = 0;

   # to make sure to check local query def only once
   $self->{CHECK_LOCAL_QUERY} = 0;

   my @args = @_;
   while (@args) {
      my $key = shift @args;
      $self->{$key} = shift @args;
   }

   bless($self, $class);
   return $self;
}

######################################################################
#::method Op_batch

=over 6

=item C<Op_batch(action, filename)>

This method provides an operation that may be invoked from the command 
line to process a list of pending ClearQuest Checkin assocations.

The two optional arguments are the action (do,undo,noop) and the batch 
filename. If not provided the default action is assumed to be "do" and 
the filename is assumed to be the value of CQCC_ASSOC_BATCHFILE.

=back

=cut

######################################################################

sub Op_batch {
    my ($self, $action, $filename) = @_;
    my $abort;
    my $fn = "TriggerCQCC::Op_batch";
    my $t1 = &TimerStart("$fn\n"); 
    my %cqHash = ();   # Hash list of defect records that have op lists
    my $init_err_count = &IOstd->{ERR_COUNT};

    #Cleanup handler will display message about batch failing
    my @cleanupHandler = IOGetCleanupHandler();
    &IOSetCleanupHandler($self, "batch_cleanup()");

    # actions: do(default), undo(reverse), noop(check file only)
    if ($action eq "") {
        $action = "do";
    }
    if ($action ne "do" && $action ne "undo" && $action ne "noop") {
        &IOMessage("FATAL", ["TCQCC_E_BATCH_UNKNOWN_OP", $action]);
    }

    # RETRIEVE DATA from batch file:
    #
    # If no explicit filename given, look for the option

    if (! defined $filename || $filename eq "") {
        $filename = $self->{BATCHFILE};
    }

    &IOLockOperation($filename, $IO_LOCKOP_NET,
                     "AUTOBREAK_MINS" => &GetParm("CM_LOCK_BREAK_MINS"),
                     "WAIT_WARN_MINS" => &GetParm("CM_LOCK_WARNING_MINS"));
    # Read in batch requests and store in hash tables for processing
    # Warn if no batch file found, may have already been processed

    my @preProcessInfo = stat($filename);  # Remember file stat pre-read

    if (! open(INFILE, $filename)) {
        my $seriesID = $ENV{"CLEARCASE_PPID"};
        # Report warning if no batch file and running from command line
        # instead of clearcase operations. Multi-threaded clearcase 
        # operations could already have posted this file. 
        if (! $seriesID || $seriesID eq "") {
            &IOMessage("WARN", ["TCQCC_E_BATCH_INPUT_FAILED", $filename]);
        }
        &IOLockOperation($filename, $IO_LOCKOP_UNLOCK);
        &IOExitHandler(0);
    }
    &IODebugPrint("BASIC", "$fn: reading CQ assoc info from $filename");

    my ($line,%opHash, $opHashRef, %oidHash);        # Temps: line buffer
    my $result = 1;
    my $lineNumber = 0;

    $self->Connect("ClearCase");

    while ($line = <INFILE>) {
        $lineNumber++;
        # Match against <CQ>defectid <op> VOlist
        chomp $line;
        $line =~ s/\r$//;   # If cross-platform remove extra line terminator
        if ($line ne "" && $line =~ /<CQ>(\S*)\s*(<[\+|\-]>)\s*(\S[\S|\s]*)$/) {
             my $defect = $1;
             my $op = $2;
             my $voList = $3;
             my $validation_result; 
             # Retrieve existing hash for this defect or start new one
             $opHashRef = $cqHash{$defect};
             if (! defined $opHashRef) { 
                 my %newHash = ();
                 $opHashRef = \%newHash;
             }
             %opHash = %{$opHashRef};
             # Add entry for voList/op pair
             # Reverse operation here for undo action
             my ($path,$vobid,$oid,$oldoid) = split(",", $voList);

             # Validate the pattern of path, vobid and oid
             $validation_result = $self->{CC}->ValidateVOList( "Batch file input line $lineNumber", $path, $vobid, $oid); 

             my $newOp = ($action ne "undo") ? $op :
                           ($op eq $addOp) ? $delOp : $addOp;

             # Only keep one operation per pathname
             # Canceling out opposite operations (CO+, CO-/CI+) 
             #    works fine in a contiguous batch but causes problems
             #    if there was a batch error and more ops were added 
             #    later so it isn't a safe thing to do. 

             $opHash{$path} = join(",",$newOp,$vobid,$oid,$oldoid);

             my %tempHash = %opHash;
             $cqHash{$defect} = \%tempHash;
             # Track different oids in use as a way of counting how many 
             # versions are involved in this batch. 

             $oidHash{$oid} = $oldoid;
        }
        else {
            # Report error reading line, avoid CQ processing later
            &IOLockOperation($filename, $IO_LOCKOP_UNLOCK);
            &IOMessage("FATAL", ["TCQCC_E_BATCH_INPUT_ERROR", $lineNumber, $line]);
            $result = 0;
        }
    }
    close INFILE;

    # If action is do or undo, run the batch, otherwise exit now
    my @cqHashKeys = keys %cqHash;
    if ($result && 
           ($action eq "do" || $action eq "undo") && 
           ($#cqHashKeys > -1)) {
        # ACT ON INFORMATION: Apply the requests to the CQ database
        my $db = &GetParm("CQCC_DATABASE");
        my $entity = &GetParm("CQCC_ENTITY");
        $self->Connect("ClearQuest");
        $result = $self->{CQ}->ProcessBatchOps(\%cqHash);
        # Restore original DB/Entity settings so user isn't surprised later
        $self->{CQ}->SetDBEntity($db,$entity);
    }

    # Move INFILE to INFILE.old - save file but 
    # automatically move it aside once it has been processed once. 

    &TimerStop($t1);
    if ($result) {
        my @postProcessInfo = stat($filename);  # Remember file stat pre-read

        # TODO: Consider dropping lock-verification in favor of size check only
        # Only rename the file if we are still the lock holder and size didn't change
        # AND if there were no errors reported during the processing of the batch
        my $time = "";
        if (&GetParm("CQCC_ASSOC_BATCH_FILE_OVERWRITE") eq "FALSE")
        {
            my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
                = localtime(time);
            $time = sprintf("_%04d%02d%02d_%02d%02d%02d", 
                $year + 1900, $mon + 1, $mday, $hour, $min, $sec);
        }

        if (&IOstd->{ERR_COUNT} == $init_err_count &&
                $preProcessInfo[7] eq $postProcessInfo[7] && 
                &IOLockOperation($filename, $IO_LOCKOP_VERIFY) eq 1) {
                    &IODebugPrint("BASIC", 
                        "$fn: Completed operation, renaming batch to $filename". ".old" . "$time");
                        rename $filename, $filename . ".old" . "$time" ;
        }
        my $confirm = &GetParm("CQCC_ASSOC_BATCH_CONFIRM");
        if ($confirm ne "OFF" ) {
            # BETTER: If $#oidKeys is 2, then check if first entries oldoid = second entry 
            # or vice versa. 
            # Also need to catch the case of 
            my @oidKeys = keys %oidHash;

            # If we are at the end of a multi-file series
            # -OR- we find more than one version in the worklist
            # then we are in a multi-file situation.

            my $multiOp = 0;
            if (&GetParm("CLEARCASE_BEGIN_SERIES") eq 0 &&
                &GetParm("CLEARCASE_END_SERIES") eq 1) {
                $multiOp = 1;

            } elsif ($#oidKeys > 0) {
                $multiOp = 1;
                # catch case of delete-checkout, add-checkin of same file
                if ($#oidKeys eq 1) {
                    my ($first, $second) = @oidKeys;
                    if ($oidHash{$first} eq $second || 
                        $oidHash{$second} eq $first) {
                        $multiOp = 0;
                    }
                }
            }

            if ($confirm eq "ALWAYS" || $multiOp) {
                &IOMessage("INFO", ["TCQCC_I_BATCH_CONFIRM"]);
            }
        }
        &IOSetCleanupHandler(@cleanupHandler);
        &IOLockOperation($filename, $IO_LOCKOP_UNLOCK);
        IOExitHandler(0);
    }
    &IOLockOperation($filename, $IO_LOCKOP_UNLOCK);
    &IOExitHandler(1);
}

sub batch_cleanup {
    my ($self) = @_;
    # Not fatal because the changes that failed are already in the batch
    # and failing a checkin or uncheckout will still leave those changes there.
    # They will be repeated again at the next change until they succeed. 
    # so don't stop pre-op uncheckouts or checkins since it is "too late"
    # and the CQ changes that go with them will be repeated later. 
    &IOMessage("ERROR", ["TCQCC_E_BATCH_OP_FAILED",$self->{BATCHFILE}]);    
    IOExitHandler(1);
}

######################################################################
#::method Op_checkin

=over 6

=item C<Op_checkin()>

This method handles the ClearCase checkin() operation. If batch 
processing is enabled, CQ association changes are written to a log
for later processing; if this is the last file in a series, then 
a batch job is automatically begun to commit changes seen so far. 
Otherwise the standard checkin operation is called. 

When called from a pre-op trigger it calls the SelectAssocs() method
to select the desired associations through user interaction or
automation during batch operations. The associations it makes may be
incorrect because the object id of the checked in file will not be
known with certainty until ClearCase completes the checkin operation.

When called from a post-op trigger it fixes the associations to use
the correct object id for the checked in file.

=back

=cut

######################################################################

sub Op_checkin {
    my $self = shift;
    my $fn = "TriggerCQCC::Op_checkin";
    
    $self->Connect("ClearCase");

    # Skip postop check unless enabled
    if ($self->{POSTOP} &&
        (&GetParm("CQCC_POSTCHECKIN_COMMIT") eq "FALSE" &&
         &GetParm("CQCC_POSTCHECKIN_VERIFY") eq "FALSE" &&
         $self->{BATCHFILE} eq "")) {
        $self->SaveCache();
        &IOExitHandler(0);
    }

    ##################################################
    # Shared setup

    if ($self->{POSTOP}) {
        $self->{CC}->SetXPN(GetParm("CLEARCASE_VOB_PN"), GetParm("CLEARCASE_XPN"));
    }
    $self->{EXISTING_ASSOCS} = $self->{CC}->ExistingAssocList();

    # Check for the database, entity pair we need
    my @entDBList = $self->{EXISTING_ASSOCS}->GetEntityDBList();
    if ($#entDBList > -1) {
        my ($entity,$db) = split(":",$entDBList[0]);
        &ForceParm("CQCC_DATABASE", $db);
        &ForceParm("CQCC_ENTITY", $entity);
    }

    ##################################################
    # Pre-op (interactive part of checkin) 

    if ($self->{POSTOP} eq 0) { 
        # If pre-op trigger and post-op trigger will fire,
        # enable series cache
        if (&GetParm("CQCC_POSTCHECKIN_COMMIT") eq "TRUE" ||
            &GetParm("CQCC_POSTCHECKIN_VERIFY") eq "TRUE") {
            &ForceParm("SAVE_SERIES_CACHE","TRUE");
        }

        # Commit_TryNoCQLogin won't return if successful
        # to avoid interactive processing in the middle of batches
        $self->Commit_TryNoCQLogin();

        $self->{VO_LIST} = $self->{CC}->GetVOList("XPN");
        $self->Connect("ClearQuest");
        $self->{DESIRED_ASSOCS}->SetAssocs(
            $self->{EXISTING_ASSOCS}->GetAssocs());
        &ForceParm("DesiredAssocList",
            join(", ", $self->{DESIRED_ASSOCS}->GetBugList()));
        $self->SelectAssocs("checkin");
        &IOExitHandler(1);
    }

    ##################################################
    # Post-Op operations: 
    # Covers situations with a combination of 
    # POST_CHECKIN Verify or Commit and batched operations
    # 
    # Truth table: 
    #     Batched   Verify   Commit:  Handled 
    #        F        F        F      (Skipped at opening)
    #        F        T        F      Verify block, immediate 
    #        F        F        T      Commit block, CQAssocs
    #        F        T        T      Verify and Commit block
    #        T        F        F      Autorun block only
    #        T        T        F      New forces to T/T/T
    #        T        F        T      Commit, Autorun blocks
    #        T        T        T      Verify/wait, Commit, Autorun
    
    else {
        # Rely on the pre-checkin cached information because 
        # we need the original checkout OID to use as "oldOID". 

        $self->{VO_LIST} = &GetParm("CACHE_VOLIST");

        # Checkin handles initialization itself so it can avoid
        # initializing ClearQuest if there is no need to use it

        ##################################################
        # POSTCHECKIN_VERIFY - fix the oids if necessary
        # Otherwise save the CC operation for performance sake.
        # Note: If Batched and Verify are true, 
        #       then Commit is forced on by new() 
        #       to simplify verify repair coding
        # If CCase cached that config spec doesn't select checkout, 
        # then force verify to ensure we have the right OID. 

        if ((&GetParm("CQCC_POSTCHECKIN_VERIFY") eq "TRUE")
            || (&GetParm("CACHE_CHECKOUT_NOT_SELECTED") eq 1)
           ) {

            my $newVOList = $self->{CC}->GetVOList("XPN");
            my ($cqname1,$vobid1,$oid1,$oldoid1) = 
                split(",", $self->{VO_LIST});
            my ($cqname2,$vobid2,$oid2,$oldoid2) = 
                split(",", $newVOList);
            if ($cqname1 ne $cqname2 || $vobid1 ne $vobid2 ||
                   $oid1 ne $oid2) {
                $self->{VO_LIST} = 
                    join(",", $cqname2, $vobid2, $oid2, $oldoid1);

                &IODebugPrint("BASIC",
                    "$fn(post): Updating oids\n" .
                    "   Original: " . &GetParm("CACHE_VOLIST") . "\n" .
                    "   New: " . $self->{VO_LIST} . "\n");

                # If not batched or post-commit, make changes immediately
                # (New() will force post-commit if verify and batched)
                # Otherwise post-checkin commit will make them correctly
                # in its first pass below

                if (&GetParm("CQCC_POSTCHECKIN_COMMIT") eq "FALSE") {
                    my $existingAssocs = $self->{EXISTING_ASSOCS};
                    my @assocs = $existingAssocs->GetAssocs();
                    if ($#assocs == -1) {
                        $self->SaveCache();
                        &IOExitHandler(0);
                    }
                    $self->Connect("ClearQuest");
                    # Delete existing associations using original OID
                    if ($self->{CQ}->EditAssocs(&GetParm("CACHE_VOLIST"),
                           undef, $existingAssocs) &&
                        $self->{CQ}->EditAssocs($self->{VO_LIST},
                           $existingAssocs, undef)) {
                        $self->SaveCache();
                        &IOExitHandler(0);
                    }
                    &IOExitHandler(1);
                }
            }
        }

        ##################################################
        # POSTCHECKIN_COMMIT - make actual CQ transactions in post-op

        my $commit_result = 1;
        if (&GetParm("CQCC_POSTCHECKIN_COMMIT") eq "TRUE") {
            $self->{DESIRED_ASSOCS}->SetAssocs(
                split(",", &GetParm("CACHE_DESIRED_ASSOCS")));
            my ($addAssocs,$delAssocs) = 
                $self->{EXISTING_ASSOCS}->CreateEditLists(
                    $self->{DESIRED_ASSOCS});
            my $ccResult = $self->{CC}->EditLinks($addAssocs, $delAssocs);
            if ($self->Commit_CQAssocs() && $ccResult) {
                &IODebugPrint("BASIC",
                    "$fn: CQCC_POSTCHECKIN_COMMIT succeeded: " . 
                    &GetParm("CACHE_DESIRED_ASSOCS") . "\n");
            } else {
                &IODebugPrint("BASIC",
                    "$fn: CQCC_POSTCHECKIN_COMMIT failed: " . 
                    &GetParm("CACHE_DESIRED_ASSOCS") . "\n");
                $commit_result = 0;
            }
         }

         ##################################################
         # Commit_TryAutoRun only succeed on PostOp at end of series
         if ($self->{BATCHFILE} ne "") {
             $self->Commit_TryAutoRun();
         }
         if ($commit_result == 0) {
             &IOExitHandler(1);
         }
     }
     $self->SaveCache();     
     &IOExitHandler(0);
}

#####################################################################
#::method Op_checkout

=over 6

=item C<Op_checkout()>

This method carries out the ClearCase checkout() operation. If batch
processing is enabled, CQ association changes are written to a log for
later processing; if this is the last file in a series, then a batch
job is automatically started to commit changes seen so far. Otherwise
it calls the SelectAssocs() function it shares in common with
checkin().  If this fails, ClearCase will be asked to uncheckout the
file.

=back

=cut

######################################################################

sub Op_checkout {
    my $self = shift;

    $self->Connect("ClearCase");

    # Commit_TryNoCQLogin won't return if successful
    $self->Commit_TryNoCQLogin();

    # If failed on checkout then uncheckout the file
    &IOSetCleanupHandler($self, "checkout_cleanup()");

    $self->Connect("ClearQuest");

    # Note: Code is a little misleading as it would normally exit
    # during SelectAssocs either through "ok" or "cancel".

    $self->{VO_LIST} = $self->{CC}->GetVOList("PN");
    my $result = $self->SelectAssocs("checkout");

    &IOSetCleanupHandler();

    return $result;
}

sub checkout_cleanup {
    my $self = shift;

    &IODebugPrint("BASIC","TriggerCQCC::checkout_cleanup " .
         "(uncheckout file)\n");
    $ENV{"CQCC_NESTED_OP"} = $self->{CC}->{XPN};

    $self->{CC}->UnCheckout($self->{CC}->{XPN});
}

######################################################################
#::method Op_edit

=over 6

=item C<Op_edit(vobPathName, XPN)>

This method provides an operation that may be invoked from the
command line to change the associations for an already checked
in pathname.

The two optional arguments are the extended pathname (XPN)
and the VOB pathname (vob_pn). If not provided they will be
prompted for.

=back

=cut

######################################################################

sub Op_edit {
    my ($self, $vob_pn, $xpn) = @_;
    my $abort; 

    # Use private IOBase to avoid conflicts with uninitialized GUI
    # GUI IOBase will be created later as the standard IOStd instance
    my $iobase = new CQCC::IOBase(); 
    $iobase->Initialize();

    $self->Connect("ClearCase");

    &ForceParm("CACHE_APPLY_TO_ALL",0);
    &ForceParm("CACHE_SERIES",0);
    &ForceParm("CQCC_QUERY_OID", "TRUE");

    if (!defined $vob_pn) {
        ($abort, $vob_pn) = $iobase->Prompt(["TCQCC_P_VOBPN"]);
    }
    if ($abort) {
        &IOExitHandler(1);
    } 
    if (!defined $xpn) {
        ($abort, $xpn) = $iobase->Prompt(["TCQCC_P_XPN"]);
    }
    if ($abort) {
        &IOExitHandler(1);
    } 

    $self->{CC}->SetXPN($vob_pn, $xpn);

    $self->{VO_LIST} = $self->{CC}->GetVOList("XPN");

    $self->{EXISTING_ASSOCS} = $self->{CC}->ExistingAssocList();
    $self->{DESIRED_ASSOCS}->SetAssocs(
        $self->{EXISTING_ASSOCS}->GetAssocs());

    # Initialize ClearQuest interface once we know we have a good XPN

    $self->Connect("ClearQuest");
    $self->SelectAssocs("edit");
    return;
}

######################################################################
#::method Op_pw

=over 6

=item C<Op_pw()>

This method is invoked to obtain the encrypted password string needed
by the cache manager scheduler task. 

=back

=cut

######################################################################

sub Op_pw {
    my $self = shift;

    # Create CQ connection object, clear login info and prompt again
    $self->{CQ} = &IOClassFactory("ClearQuestInterface");
    $self->{CQ}->ClearLogonParms();
    $self->{CQ}->Connect();

    # Recover CQ login information and present in cache manager form
    my $parmfile = $self->{CQ}->{PARMFILE};
    my ($answer) = &IOFileRead($parmfile,0);
    my ($user, $pass, $cqusername) = split(" ", $answer);
    print "$user:$pass\n";
    return 0;
}

######################################################################
#::method Op_test

=over 6

=item C<Op_test()>

This method is invoked to carry out the -test mode operation. It will
ask the ClearQuest object to do a basic query. Testing automatically
enables debugging and tracing.

=back

=cut

######################################################################

sub Op_test {
    my $self = shift;

    #Confirm schema

    #Do sample query

    $self->Connect("ClearCase");
    $self->Connect("ClearQuest");

    my @bug_query =
        $self->{CQ}->PerformQuery(&GetParm("CQCC_DATABASE"),
                                  &GetParm("CQCC_ENTITY"),
                                  undef, "STANDARD", "STANDARD",
                                  "FALSE");
    my $bugline;
    for $bugline (@bug_query) {
        print $bugline,"\n";
    }

    return 0;
}

######################################################################
#::method Op_uncheckout

=over 6

=item C<Op_uncheckout()>

This method carries out the ClearCase uncheckout operation by deleting
existing associations on checked out versions before they are
unchecked out. If the batch file option is enabled, it writes the
necessary deletion operations to the file for later use.

=back

=cut

######################################################################

sub Op_uncheckout {
    my $self = shift;

    $self->Connect("ClearCase");

    #Check for nested trigger and skip if found
    if ($ENV{"CQCC_NESTED_OP"} eq &GetParm("CLEARCASE_XPN")) {
        &IODebugPrint("DETAILED",
                      "Op_uncheckout - nested operation skipped\n");
        &IOExitHandler(0);   # Normal return expected
    } 

    # Does subset of what Commit_TryNoCQLogin does

    $self->{VO_LIST} = $self->{CC}->GetVOList("PN");

    $self->{EXISTING_ASSOCS} = $self->{CC}->ExistingAssocList();

    # Clear DESIRED_ASSOCS - uncheckout doesn't require any
    $self->{DESIRED_ASSOCS}->SetAssocs(());

    return $self->Commit_Assocs();
}

######################################################################
#::method QueryNameList

=over 6

=item C<@list = QueryNameList(type)>

This method returns the acceptable list of possible queries. 

If the type argument is "LOCAL" then the list will include any
valid LOCAL queries and may include ClearQuest queries if the list
is already cached. Otherwise the full list will be provided. This
optimization allows validating local queries without spending time
getting the entire list from the ClearQuest database. 

=back

=cut

######################################################################

sub QueryNameList {
    my ($self, $type) = @_;

    my $queryEnable = &GetParm("CQCC_QUERY_ENABLE");
    my @selectQueryList = $self->{QUERY_NAME_LIST} ? 
                          @{ $self->{QUERY_NAME_LIST} } : ();

    # Recover existing list and reuse it within session only.
    # If existing list empty then reload it.

    if (! ( @selectQueryList && $#selectQueryList > -1)) {
       if (($type ne "LOCAL") && 
           ($queryEnable eq "CQ" || $queryEnable eq "BOTH")) {
           if ($perl5006 && $self->{CQ}->can("GetQueryList")) {
               @selectQueryList =
                  sort $self->{CQ}->GetQueryList("BOTH");
               # Apply filter 
               my $filter = &GetParm("CQCC_QUERY_FILTER");
               if ($filter) { 
                  my @newList;
                  for (@selectQueryList) { 
                     if ($_ =~ $filter) { 
                         push @newList, $_;
                     }
                  }
                  @selectQueryList = @newList;
               }
           }
           # If it can't do it then fall back to LOCAL mode (for web)
           else {
               &ForceParm("CQCC_QUERY_ENABLE", "LOCAL");
           }
       }
       my $schema = 
          &GetCQSchemaByEntity(&GetParm("CQCC_ENTITY"));
       if ($queryEnable eq "LOCAL" || $queryEnable eq "BOTH") {
           my @localQueries = $schema->GetQueryList();
           for (@localQueries) { 
              push @selectQueryList, "LOCAL/$_";
           }
           @selectQueryList = sort @selectQueryList;
       }
       if ($type ne "LOCAL") { 
           $self->{QUERY_NAME_LIST} = \@selectQueryList;
       }
    }

    return @selectQueryList;
}

######################################################################
#::method QueryNameValidate

=over 6

=item C<QueryNameValidate(option)>

This method confirms that the current Query name is valid for 
the current database. If not, it is reset to LOCAL/STANDARD.

If the option argument is REFRESH, the cached query name list
will be cleared to make sure the list is current. 

Returns 1 if the current query name was valid, 0 if it was not
and had to be reset to LOCAL/STANDARD.

=back

=cut

######################################################################

sub QueryNameValidate {
    my ($self, $option) = @_;
    my $queryName = &GetParm("CQCC_QUERY_NAME");
    my $queryType = &GetParm("CQCC_QUERY_TYPE");
    my $queryPath = ($queryType eq "CQ") ? $queryName :
                    "LOCAL/$queryName";
    if ($option eq "REFRESH") { 
        $self->{QUERY_NAME_LIST} = undef;
    } 
    my @selectQueryList = $self->QueryNameList($queryType);

    for (@selectQueryList) {
       if ($queryPath eq $_) {
          return 1;
       }
    }    
    &ForceParm("CQCC_QUERY_NAME", "STANDARD");
    &ForceParm("CQCC_QUERY_TYPE", "LOCAL");
    return 0;
}

######################################################################
#::method ProcessAutoAssocEVs

=over 6

=item C<ProcessAutoAssocEVs(reportErrors)>

This internal method processes environment variables dealing with
automated associations such as CQCC_AUTO_ASSOCIATE and
CQCC_COMMENT_PATTERN.  When an appropriate environment variable is
recognized, the method sets the DESIRED_ASSOCS list and returns 1.

If reportErrors is 0, errors are not reported to avoid multiple
reports from multiple calls. 

=back

=cut

######################################################################

sub ProcessAutoAssocEVs {
    my ($self, $reportErrors) = @_;
    my $result = 0;
    my $fn = "TriggerCQCC::ProcessAutoAssocEVs";
    my $applyToAll = &GetParm("CACHE_APPLY_TO_ALL");
    &ForceParm("CACHE_VOLIST", $self->{VO_LIST});

    ########################################
    # Process Environment variable 

    my $override = &GetParm("CQCC_AUTO_ASSOCIATE");
    # Strip off any enclosing quotes
    if ($override =~ /\"(.*)\"/) {
        $override = $1;
    }

    if ((&GetParm("CQCC_AUTO_ASSOCIATE_ENABLE") eq "TRUE") &&
        (defined($override) && $override ne "")) {

       IODebugPrint("BASIC", "$fn: \n" .
           "Using defects from CQCC_AUTO_ASSOCIATE ($override)\n");

       if ($self->{DESIRED_ASSOCS}->ProcessBugString(
                &GetParm("CQCC_ENTITY"), 
                ($override eq "-" ? "" : $override))) {
            $result = 1;
       }
       else {
           # Report parsing error - CUSTOMIZE for source
           if ($reportErrors) {
               &IOMessage("INFO", 
                          ["TCQCC_E_BUGLIST_INPUT", 
                          $override . " (CQCC_AUTO_ASSOCIATE)",
                          &GetParm("CQCC_DATABASE"),
                          &GetParm("CQCC_ENTITY") ]);
           }
       }
    }
            
    ########################################
    # Process Comment Pattern option
 
    else {
        my $pattern = &GetParm("CQCC_COMMENT_PATTERN");
        my $comment = &GetParm("CLEARCASE_COMMENT"); 
        if (defined($pattern) && $pattern ne "" &&
            defined($comment) && $comment ne "" &&
            $comment =~ $pattern && $1 ne "") {
            my $match = $1;
            IODebugPrint("BASIC", "$fn: \n" .
                "Using defects from CQCC_COMMENT ($match)\n");

            if ($self->{DESIRED_ASSOCS}->ProcessBugString(
                    &GetParm("CQCC_ENTITY"), 
                    ($match eq "-" ? "" : $match))) {
                $result = 1;
            }
            else {
                # Report parsing error - CUSTOMIZE for source
                if ($reportErrors) {
                    &IOMessage("INFO", 
                               ["TCQCC_E_BUGLIST_INPUT", 
                                 $match . " (CLEARCASE_COMMENT)", 
                                 &GetParm("CQCC_DATABASE"),
                                 &GetParm("CQCC_ENTITY") ]);
                }
            }
        }
    }

    ########################################
    # Process checkins for checkout only trigger installations
    # By using the checkout associations and updating them

    if (&GetParm("SILENT") eq "TRUE" && $result eq 0) {
        # Check DESIRED_ASSOCS is defaulted to the checkout assoclist 
        IODebugPrint("BASIC", "$fn: \n" .
            "Using defects from Checkout because Checkout only trigger\n" .
            "installed in VOB with silent Checkin trigger\n");
        $self->{DESIRED_ASSOCS}->SetAssocs(
            $self->{EXISTING_ASSOCS}->GetAssocs());
        $result = 1;
    }

    return $result;
}

######################################################################
#::method SelectAssocs

=over 6

=item C<SelectAssocs(operation)>

This method provides shared functionality for a number of ClearCase
operations. It sets up the initial settings appropriate for the
operation then invokes the main menu which interacts with the user to
display the current assocations, make changes, and update the desired
and existing association lists.

It also processes the ApplyToAll settings from the last session to
automate processing versions in a batch. When ApplyToAll is 1, then
the cached associations from the last session will override the
checkout associations if any. When ApplyToAll is 2, then the checked
out associations for each version are used when that version is
checked in.

=back

=cut

######################################################################

sub SelectAssocs {
    my ($self,$op) = @_;

    my $initialChoice = "";
    my $applyToAll = &GetParm("CACHE_APPLY_TO_ALL");

    &ForceParm("CACHE_VOLIST", $self->{VO_LIST});

    ##################################################
    # If we are still processing the same batch then
    # try to commit using the cached defect list.
    # If there are problems, they will be reported and user will be
    # returned to the main menu to find out what they want to do.

    my $series = &GetParm("CLEARCASE_SERIES_ID");
    if (defined $series && $series ne "" &&
            &GetParm("CACHE_SERIES") eq $series &&
            $applyToAll > 0) {
        # Note: _ApplyToAll msg tag takes and argument but the 
        # matching operation doesn't so don't include it here. 
        $initialChoice = [ "Action_Commit_ApplyToAll" ];

        # If ApplyToAll=1 then override checkedout assocs with
        # the cached values

        if ($applyToAll == 1) {
           $self->{DESIRED_ASSOCS}->SetAssocs(
               split(",", &GetParm("CACHE_DESIRED_ASSOCS")));

        # Otherwise if it is 2 then use the checked in assocs
        } elsif ($applyToAll == 2) {
           $initialChoice =
               [ "Action_Commit_ApplyToAll_UseCheckouts" ];
        } elsif ($applyToAll == 3) {
           $initialChoice = [ "Action_Cancel_ApplyToAll" ];
        }

    } else {
        &ForceParm("CACHE_APPLY_TO_ALL", 0);
        $applyToAll = 0;
    }

    ########################################
    # Process Environment variable 

    if ($initialChoice eq "" && $self->ProcessAutoAssocEVs(1) eq 1) {
        $initialChoice = [ "Action_Commit" ];
    } 

    ########################################
    # Retrieve and use cached value for bug list

    my @desiredList = $self->{DESIRED_ASSOCS}->GetBugList();
    &ForceParm("DesiredAssocList", join(", ", @desiredList));

    ########################################
    # Provide debug output 

    if ($applyToAll > 0) {
       &IODebugPrint("BASIC",
           "TriggerCQCC::SelectAssocs: " .
               "Batch processing enabled (ApplyToAll = " .
           $applyToAll . ")\n    " .
           ($applyToAll == 1 ?
              "1 = Apply chosen associations for all in batch \n" :
              ($applyToAll == 2 ?
              "2 = Use checkout associations for each in batch \n" :
              "3 = Cancel operation against rest of batch\n")) .
           "    Using " . join (", ", @desiredList));
    }

    if ($initialChoice ne "") {
        my ($choice) = @{$initialChoice};
        my $cmd = "\$self->$choice()";
        eval $cmd;
        # if it returns then fall into main menu
    }

    ########################################
    # If allowed and possible, use GUI

    $self->QueryNameValidate("REFRESH"); # Validate query before use

    if (&GetParm("CQCC_GUI_ENABLE") ne "OFF" &&
        $perl5006 && &IOstd()->can("GetMainWindow")) {
        $self->GUIDisplay();
    }
    else {
        &CheckLocalQuery();
        $self->MenuDisplay("MENU_ASSOCIATION", undef);
    }
    return;  # Expect to exit through Menu
}

######################################################################
#::Tail
######################################################################

1;

